import scipy.signal

from tsfel.feature_extraction.features_utils import *


# ############################################# TEMPORAL DOMAIN ##################################################### #


@set_domain("domain", "temporal")
def autocorr(signal):
    """Computes autocorrelation of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which autocorrelation is computed

    Returns
    -------
    float
        Cross correlation of 1-dimensional sequence

    """

    return float(np.correlate(signal, signal))


@set_domain("domain", "temporal")
def calc_centroid(signal, fs):
    """Computes the centroid along the time axis.

    Parameters
    ----------
    signal : nd-array
        Input from which centroid is computed
    fs: int
        Signal sampling frequency

    Returns
    -------
    float
        Temporal centroid

    """

    time = compute_time(signal, fs)

    energy = np.array(signal) ** 2

    t_energy = np.dot(np.array(time), np.array(energy))
    energy_sum = np.sum(energy)

    if energy_sum == 0 or t_energy == 0:
        centroid = 0
    else:
        centroid = t_energy / energy_sum

    return centroid


@set_domain("domain", "temporal")
def minpeaks(signal):
    """Computes number of minimum peaks of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which minimum number of peaks is counted
    Returns
    -------
    float
        Minimum number of peaks

    """
    diff_sig = np.diff(signal)

    return np.sum([1 for nd in range(len(diff_sig[:-1])) if ((diff_sig[nd] < 0) and (diff_sig[nd + 1] > 0))])


@set_domain("domain", "temporal")
def maxpeaks(signal):
    """Computes number of maximum peaks of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which maximum number of peaks is counted

    Returns
    -------
    float
        Maximum number of peaks

    """
    diff_sig = np.diff(signal)

    return np.sum([1 for nd in range(len(diff_sig[:-1])) if ((diff_sig[nd + 1] < 0) and (diff_sig[nd] > 0))])


@set_domain("domain", "temporal")
def mean_abs_diff(signal):
    """Computes mean absolute differences of the signal.

   Parameters
   ----------
   signal : nd-array
       Input from which mean absolute deviation is computed

   Returns
   -------
   float
       Mean absolute difference result

   """

    return np.mean(np.abs(np.diff(signal)))


@set_domain("domain", "temporal")
def mean_diff(signal):
    """Computes mean of differences of the signal.

   Parameters
   ----------
   signal : nd-array
       Input from which mean of differences is computed

   Returns
   -------
   float
       Mean difference result

   """

    return np.mean(np.diff(signal))


@set_domain("domain", "temporal")
def median_abs_diff(signal):
    """Computes median absolute differences of the signal.

   Parameters
   ----------
   signal : nd-array
       Input from which median absolute difference is computed

   Returns
   -------
   float
       Median absolute difference result

   """

    return np.median(np.abs(np.diff(signal)))


@set_domain("domain", "temporal")
def median_diff(signal):
    """Computes median of differences of the signal.

   Parameters
   ----------
   signal : nd-array
       Input from which median of differences is computed

   Returns
   -------
   float
       Median difference result

   """

    return np.median(np.diff(signal))


@set_domain("domain", "temporal")
def distance(signal):
    """Computes signal traveled distance.

    Calculates the total distance traveled by the signal
    using the hipotenusa between 2 datapoints.

    Parameters
    ----------
    signal : nd-array
        Input from which distance is computed

    Returns
    -------
    float
        Signal distance

    """
    diff_sig = np.diff(signal)
    return np.sum([np.sqrt(1 + df ** 2) for df in diff_sig])


@set_domain("domain", "temporal")
def sum_abs_diff(signal):
    """Computes sum of absolute differences of the signal.

   Parameters
   ----------
   signal : nd-array
       Input from which sum absolute difference is computed

   Returns
   -------
   float
       Sum absolute difference result

   """

    return np.sum(np.abs(np.diff(signal)))


@set_domain("domain", "temporal")
def zero_cross(signal):
    """Computes Zero-crossing rate of the signal.

    Corresponds to the total number of times that the signal changes from
    positive to negative or vice versa.

    Parameters
    ----------
    signal : nd-array
        Input from which the zero-crossing rate are computed

    Returns
    -------
    int
        Number of times that signal value cross the zero axis

    """

    return len(np.where(np.diff(np.sign(signal)))[0])


@set_domain("domain", "temporal")
def total_energy(signal, fs):
    """Computes the total energy of the signal.

    Parameters
    ----------
    signal : nd-array
        Signal from which total energy is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Total energy

    """

    time = compute_time(signal, fs)

    return np.sum(np.array(signal) ** 2) / (time[-1] - time[0])


@set_domain("domain", "temporal")
def slope(signal):
    """Computes the slope of the signal.

    Slope is computed by fitting a linear equation to the observed data.

    Parameters
    ----------
    signal : nd-array
        Input from which linear equation is computed

    Returns
    -------
    float
        Slope

    """
    t = np.linspace(0, len(signal) - 1, len(signal))

    return np.polyfit(t, signal, 1)[0]


@set_domain("domain", "temporal")
def auc(signal, fs):
    """Computes the area under the curve of the signal computed with trapezoid rule.

    Parameters
    ----------
    signal : nd-array
        Input from which the area under the curve is computed
    fs : int
        Sampling Frequency
    Returns
    -------
    float
        The area under the curve value

    """
    t = compute_time(signal, fs)

    return np.sum(np.diff(t) * signal[:-1] + signal[1:] / 2)


@set_domain("domain", "temporal")
def abs_energy(signal):
    """Computes the absolute energy of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which the area under the curve is computed

    Returns
    -------
    float
        Absolute energy

    """

    return np.sum(signal ** 2)


@set_domain("domain", "temporal")
def pk_pk_distance(signal):
    """Computes the peak to peak distance.

    Parameters
    ----------
    signal : nd-array
        Input from which the area under the curve is computed

    Returns
    -------
    float
        peak to peak distance

    """

    return np.abs(np.max(signal) - np.min(signal))


@set_domain("domain", "temporal")
def entropy(signal, prob='kde'):
    """Computes the entropy of the signal using the Shannon Entropy.

    Description in Article:
    Regularities Unseen, Randomness Observed: Levels of Entropy Convergence
    Authors: Crutchfield J. Feldman David

    Parameters
    ----------
    signal : nd-array
        Input from which entropy is computed
    prob : string
        Probability function (kde or gaussian functions are available)

    Returns
    -------
    float
        The normalized entropy value

    """

    if prob == 'kde':
        p = kde(signal)
    elif prob == 'gauss':
        p = gaussian(signal)

    if np.sum(p) == 0:
        return 0.0

    # Handling zero probability values
    p = p[np.where(p != 0)]

    if np.sum(p * np.log2(p)) / np.log2(len(signal)) == 0:
        return 0.0
    else:
        return - np.sum(p * np.log2(p)) / np.log2(len(signal))


# ############################################ STATISTICAL DOMAIN #################################################### #


@set_domain("domain", "statistical")
def hist(signal, nbins=10, r=1):
    """Computes histogram of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from histogram is computed
    nbins : int
        The number of equal-width bins in the given range
    r : float
        The lower(-r) and upper(r) range of the bins

    Returns
    -------
    nd-array
        The values of the histogram

    """

    histsig, bin_edges = np.histogram(signal, bins=nbins, range=[-r, r])  # TODO:subsampling parameter

    return tuple(histsig)


@set_domain("domain", "statistical")
def interq_range(signal):
    """Computes interquartile range of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which interquartile range is computed

    Returns
    -------
    float
        Interquartile range result

    """

    return np.percentile(signal, 75) - np.percentile(signal, 25)


@set_domain("domain", "statistical")
def kurtosis(signal):
    """Computes kurtosis of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which kurtosis is computed

    Returns
    -------
    float
        Kurtosis result

    """

    return scipy.stats.kurtosis(signal)


@set_domain("domain", "statistical")
def skewness(signal):
    """Computes skewness of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which skewness is computed

    Returns
    -------
    int
        Skewness result

    """

    return scipy.stats.skew(signal)


@set_domain("domain", "statistical")
def calc_max(signal):
    """Computes the maximum value of the signal.

    Parameters
    ----------
    signal : nd-array
       Input from which max is computed

    Returns
    -------
    float
        Maximum result

    """

    return np.max(signal)


@set_domain("domain", "statistical")
def calc_min(signal):
    """Computes the minimum value of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which min is computed

    Returns
    -------
    float
        Minimum result

    """

    return np.min(signal)


@set_domain("domain", "statistical")
def calc_mean(signal):
    """Computes mean value of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which mean is computed.

    Returns
    -------
    float
        Mean result

    """

    return np.mean(signal)


@set_domain("domain", "statistical")
def calc_median(signal):
    """Computes median of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which median is computed

    Returns
    -------
    float
        Median result

    """
    return np.median(signal)


@set_domain("domain", "statistical")
def mean_abs_deviation(signal):
    """Computes mean absolute deviation of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which mean absolute deviation is computed

    Returns
    -------
    float
        Mean absolute deviation result

    """

    return np.mean(np.abs(signal - np.mean(signal, axis=0)), axis=0)


@set_domain("domain", "statistical")
def median_abs_deviation(signal):
    """Computes median absolute deviation of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which median absolute deviation is computed

    Returns
    -------
    float
        Mean absolute deviation result

    """

    return scipy.stats.median_absolute_deviation(signal, scale=1)


@set_domain("domain", "statistical")
def rms(signal):
    """Computes root mean square of the signal.

    Square root of the arithmetic mean (average) of the squares of the original values.

    Parameters
    ----------
    signal : nd-array
        Input from which root mean square is computed

    Returns
    -------
    float
        Root mean square

    """

    return np.sqrt(np.sum(np.array(signal) ** 2) / len(signal))


@set_domain("domain", "statistical")
def calc_std(signal):
    """Computes standard deviation (std) of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which std is computed

    Returns
    -------
    float
        Standard deviation result

    """

    return np.std(signal)


@set_domain("domain", "statistical")
def calc_var(signal):
    """Computes variance of the signal.

    Parameters
    ----------
    signal : nd-array
       Input from which var is computed

    Returns
    -------
    float
        Variance result

    """

    return np.var(signal)


@set_domain("domain", "statistical")
def ecdf(signal, d=10):
    """Computes the values of ECDF (empirical cumulative distribution function) along the time axis.

    Parameters
    ----------
    signal : nd-array
        Input from which ECDF is computed
    d: integer
        Number of ECDF values to return

    Returns
    -------
    float
        The values of the ECDF along the time axis
    """
    _, y = calc_ecdf(signal)
    if len(signal) <= d:
        return tuple(y)
    else:
        return tuple(y[:d])


@set_domain("domain", "statistical")
def ecdf_slope(signal, p_init=0.5, p_end=0.75):
    """Computes the slope of the ECDF between two percentiles.

    Parameters
    ----------
    signal : nd-array
        Input from which ECDF is computed
    p_init : float
        Initial percentile
    p_end : float
        End percentile

    Returns
    -------
    float
        The slope of the ECDF between two percentiles
    """
    # check if signal is constant
    if np.sum(np.diff(signal)) == 0:
        return np.inf
    else:
        x_init, x_end = ecdf_percentile(signal, percentile=[p_init, p_end])
        return (p_end - p_init) / (x_end - x_init)


@set_domain("domain", "statistical")
def ecdf_percentile(signal, percentile=None):
    """Determines the percentile value of the ECDF.

    Parameters
    ----------
    signal : nd-array
        Input from which ECDF is computed
    percentile: list
        Percentile value to be computed

    Returns
    -------
    float
        The input value(s) of the ECDF
    """
    if percentile is None:
        percentile = [0.2, 0.8]
    if type(percentile) in [float, int]:
        percentile = [percentile]

    # calculate ecdf
    x, y = calc_ecdf(signal)

    if len(percentile) > 1:
        # check if signal is constant
        if np.sum(np.diff(signal)) == 0:
            return tuple(np.repeat(signal[0], len(percentile)))
        else:
            return tuple([x[y <= p].max() for p in percentile])
    else:
        # check if signal is constant
        if np.sum(np.diff(signal)) == 0:
            return signal[0]
        else:
            return x[y <= percentile].max()


@set_domain("domain", "statistical")
def ecdf_percentile_count(signal, percentile=None):
    """Determines the cumulative sum of samples that are less than the percentile.

    Parameters
    ----------
    signal : nd-array
        Input from which ECDF is computed
    percentile: list
        Percentile threshold

    Returns
    -------
    float
        The cumulative sum of samples
    """
    if percentile is None:
        percentile = [0.2, 0.8]
    if type(percentile) in [float, int]:
        percentile = [percentile]

    # calculate ecdf
    x, y = calc_ecdf(signal)

    if len(percentile) > 1:
        # check if signal is constant
        if np.sum(np.diff(signal)) == 0:
            return tuple(np.repeat(signal[0], len(percentile)))
        else:
            return tuple([x[y <= p].shape[0] for p in percentile])
    else:
        # check if signal is constant
        if np.sum(np.diff(signal)) == 0:
            return signal[0]
        else:
            return x[y <= percentile].shape[0]


# ############################################## SPECTRAL DOMAIN ##################################################### #

@set_domain("domain", "spectral")
def spectral_distance(signal, fs):
    """Computes the signal spectral distance.

    Distance of the signal's cumulative sum of the FFT elements to
    the respective linear regression.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral distance is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        spectral distance

    """

    f, fmag = calc_fft(signal, fs)

    cum_fmag = np.cumsum(fmag)

    # Computing the linear regression
    points_y = np.linspace(0, cum_fmag[-1], len(cum_fmag))

    return np.sum(points_y - cum_fmag)


@set_domain("domain", "spectral")
def fundamental_frequency(signal, fs):
    """Computes fundamental frequency of the signal.

    The fundamental frequency integer multiple best explain
    the content of the signal spectrum.

    Parameters
    ----------
    signal : nd-array
        Input from which fundamental frequency is computed
    fs : int
        Sampling frequency

    Returns
    -------
    f0: float
       Predominant frequency of the signal

    """

    signal = signal - np.mean(signal)
    f, fmag = calc_fft(signal, fs)

    # Condition for offset removal, since the offset generates a peak at frequency zero
    try:
        # With 0.1 the offset frequency is discarded
        cond = np.where(f > 0.1)[0][0]
    except IndexError:
        cond = 0

    # Finding big peaks, not considering noise peaks with low amplitude
    bp = scipy.signal.find_peaks(fmag[cond:], threshold=10)[0]
    if not list(bp):
        f0 = 0
    else:
        # add cond for correcting the indices position
        bp = bp + cond
        # f0 is the minimum big peak frequency
        f0 = f[min(bp)]

    return f0


@set_domain("domain", "spectral")
def max_power_spectrum(signal, fs):
    """Computes maximum power spectrum density of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which maximum power spectrum is computed
    fs : scalar
        Sampling frequency

    Returns
    -------
    nd-array
        Max value of the power spectrum density

    """

    if np.std(signal) == 0:
        return float(max(scipy.signal.welch(signal, int(fs), nperseg=len(signal))[1]))
    else:
        return float(max(scipy.signal.welch(signal / np.std(signal), int(fs), nperseg=len(signal))[1]))


@set_domain("domain", "spectral")
def max_frequency(signal, fs):
    """Computes maximum frequency of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which maximum frequency is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        0.95 of maximum frequency using cumsum
    """

    f, fmag = calc_fft(signal, fs)
    cum_fmag = np.cumsum(fmag)

    try:
        ind_mag = np.where(cum_fmag > cum_fmag[-1] * 0.95)[0][0]
    except IndexError:
        ind_mag = np.argmax(cum_fmag)

    return f[ind_mag]


@set_domain("domain", "spectral")
def median_frequency(signal, fs):
    """Computes median frequency of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which median frequency is computed
    fs: int
        Sampling frequency

    Returns
    -------
    f_median : int
       0.50 of maximum frequency using cumsum.
    """

    f, fmag = calc_fft(signal, fs)
    cum_fmag = np.cumsum(fmag)
    try:
        ind_mag = np.where(cum_fmag > cum_fmag[-1] * 0.50)[0][0]
    except IndexError:
        ind_mag = np.argmax(cum_fmag)
    f_median = f[ind_mag]

    return f_median


@set_domain("domain", "spectral")
def spectral_centroid(signal, fs):
    """Barycenter of the spectrum.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral centroid is computed
    fs: int
        Sampling frequency

    Returns
    -------
    float
        Centroid

    """

    f, fmag = calc_fft(signal, fs)
    if not np.sum(fmag):
        return 0
    else:
        return np.dot(f, fmag / np.sum(fmag))


@set_domain("domain", "spectral")
def spectral_decrease(signal, fs):
    """Represents the amount of decreasing of the spectra amplitude.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral decrease is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral decrease

    """

    f, fmag = calc_fft(signal, fs)

    fmag_band = fmag[1:]
    len_fmag_band = np.arange(2, len(fmag) + 1)

    # Sum of numerator
    soma_num = np.sum((fmag_band - fmag[0]) / (len_fmag_band - 1), axis=0)

    if not np.sum(fmag_band):
        return 0
    else:
        # Sum of denominator
        soma_den = 1 / np.sum(fmag_band)

        # Spectral decrease computing
        return soma_den * soma_num


@set_domain("domain", "spectral")
def spectral_kurtosis(signal, fs):
    """Measures the flatness of a distribution around its mean value.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral kurtosis is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral Kurtosis

    """

    f, fmag = calc_fft(signal, fs)
    if not spectral_spread(signal, fs):
        return 0
    else:
        spect_kurt = ((f - spectral_centroid(signal, fs)) ** 4) * (fmag / np.sum(fmag))
        return np.sum(spect_kurt) / (spectral_spread(signal, fs) ** 4)


@set_domain("domain", "spectral")
def spectral_skewness(signal, fs):
    """Measures the asymmetry of a distribution around its mean value.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral skewness is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral Skewness

    """

    f, fmag = calc_fft(signal, fs)
    spect_centr = spectral_centroid(signal, fs)

    if not spectral_spread(signal, fs):
        return 0
    else:
        skew = ((f - spect_centr) ** 3) * (fmag / np.sum(fmag))
        return np.sum(skew) / (spectral_spread(signal, fs) ** 3)


@set_domain("domain", "spectral")
def spectral_spread(signal, fs):
    """Measures the spread of the spectrum around its mean value.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral spread is computed.
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral Spread

    """

    f, fmag = calc_fft(signal, fs)
    spect_centroid = spectral_centroid(signal, fs)

    if not np.sum(fmag):
        return 0
    else:
        return np.dot(((f - spect_centroid) ** 2), (fmag / np.sum(fmag))) ** 0.5


@set_domain("domain", "spectral")
def spectral_slope(signal, fs):
    """Computes the spectral slope.

    Spectral slope is computed by finding constants m and b of the function aFFT = mf + b, obtained by linear regression
    of the spectral amplitude.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral slope is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral Slope

    """

    f, fmag = calc_fft(signal, fs)

    if not (list(f)) or (np.sum(fmag) == 0):
        return 0
    else:
        if not (len(f) * np.dot(f, f) - np.sum(f) ** 2):
            return 0
        else:
            num_ = (1 / np.sum(fmag)) * (len(f) * np.dot(f, fmag) - np.sum(f) * np.sum(fmag))
            denom_ = (len(f) * np.dot(f, f) - np.sum(f) ** 2)
            return num_ / denom_


@set_domain("domain", "spectral")
def spectral_variation(signal, fs):
    """Computes the amount of variation of the spectrum along time.

    Spectral variation is computed from the normalized cross-correlation between two consecutive amplitude spectra.

    Description and formula in Article:
    The Timbre Toolbox: Extracting audio descriptors from musicalsignals
    Authors Peeters G., Giordano B., Misdariis P., McAdams S.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral variation is computed.
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral Variation

    """

    f, fmag = calc_fft(signal, fs)

    sum1 = np.sum(np.array(fmag)[:-1] * np.array(fmag)[1:])
    sum2 = np.sum(np.array(fmag)[1:] ** 2)
    sum3 = np.sum(np.array(fmag)[:-1] ** 2)

    if not sum2 or not sum3:
        variation = 1
    else:
        variation = 1 - (sum1 / ((sum2 ** 0.5) * (sum3 ** 0.5)))

    return variation


@set_domain("domain", "spectral")
def spectral_maxpeaks(signal, fs):
    """Computes number of maximum spectral peaks of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which the number of maximum spectral peaks is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Total number of maximum spectral peaks

    """

    f, fmag = calc_fft(signal, fs)
    diff_sig = np.diff(fmag)

    return np.sum([1 for nd in range(len(diff_sig[:-1])) if ((diff_sig[nd + 1] < 0) and (diff_sig[nd] > 0))])


@set_domain("domain", "spectral")
def spectral_roll_off(signal, fs):
    """Computes the spectral roll-off of the signal.

    The spectral roll-off corresponds to the frequency where 95% of the signal magnitude is contained
    below of this value.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral roll-off is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral roll-off

    """

    f, fmag = calc_fft(signal, fs)
    cum_ff = np.cumsum(fmag)
    value = 0.95 * (np.sum(fmag))

    return f[np.where(cum_ff >= value)[0][0]]


@set_domain("domain", "spectral")
def spectral_roll_on(signal, fs):
    """Computes the spectral roll-on of the signal.

    The spectral roll-on corresponds to the frequency where 5% of the signal magnitude is contained
    below of this value.

    Parameters
    ----------
    signal : nd-array
        Signal from which spectral roll-on is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Spectral roll-on

    """

    f, fmag = calc_fft(signal, fs)
    cum_ff = np.cumsum(fmag)
    value = 0.05 * (np.sum(fmag))

    return f[np.where(cum_ff >= value)[0][0]]


@set_domain("domain", "spectral")
def human_range_energy(signal, fs):
    """Computes the human range energy ratio.

    The human range energy ratio is given by the ratio between the energy
    in frequency 0.6-2.5Hz and the whole energy band.

    Parameters
    ----------
    signal : nd-array
        Signal from which human range energy ratio is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Human range energy ratio

    """

    f, fmag = calc_fft(signal, fs)

    allenergy = np.sum(fmag ** 2)

    if allenergy == 0:
        # For handling the occurrence of Nan values
        return 0.0

    hr_energy = np.sum(fmag[np.argmin(np.abs(0.6 - f)):np.argmin(np.abs(2.5 - f))] ** 2)

    ratio = hr_energy / allenergy

    return ratio


@set_domain("domain", "spectral")
def mfcc(signal, fs, pre_emphasis=0.97, nfft=512, nfilt=40, num_ceps=12, cep_lifter=22):
    """Computes the MEL cepstral coefficients.

    It provides the information about the power in each frequency band.

    Implementation details and description on:
    https://www.kaggle.com/ilyamich/mfcc-implementation-and-tutorial
    https://haythamfayek.com/2016/04/21/speech-processing-for-machine-learning.html#fnref:1

    Parameters
    ----------
    signal : nd-array
        Input from which MEL coefficients is computed
    fs : int
        Sampling frequency
    pre_emphasis : float
        Pre-emphasis coefficient for pre-emphasis filter application
    nfft : int
        Number of points of fft
    nfilt : int
        Number of filters
    num_ceps: int
        Number of cepstral coefficients
    cep_lifter: int
        Filter length

    Returns
    -------
    nd-array
        MEL cepstral coefficients

    """

    filter_banks = filterbank(signal, fs, pre_emphasis, nfft, nfilt)

    mel_coeff = scipy.fft.dct(filter_banks, type=2, axis=0, norm='ortho')[1:(num_ceps + 1)]  # Keep 2-13

    mel_coeff -= (np.mean(mel_coeff, axis=0) + 1e-8)

    # liftering
    ncoeff = len(mel_coeff)
    n = np.arange(ncoeff)
    lift = 1 + (cep_lifter / 2) * np.sin(np.pi * n / cep_lifter)  # cep_lifter = 22 from python_speech_features library

    mel_coeff *= lift

    return tuple(mel_coeff)


@set_domain("domain", "spectral")
def power_bandwidth(signal, fs):
    """Computes power spectrum density bandwidth of the signal.

    It corresponds to the width of the frequency band in which 95% of its power is located.

    Description in article:
    Power Spectrum and Bandwidth Ulf Henriksson, 2003 Translated by Mikael Olofsson, 2005

    Parameters
    ----------
    signal : nd-array
        Input from which the power bandwidth computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        Occupied power in bandwidth

    """

    # Computing the power spectrum density
    if np.std(signal) == 0:
        freq, power = scipy.signal.welch(signal, fs, nperseg=len(signal))
    else:
        freq, power = scipy.signal.welch(signal / np.std(signal), fs, nperseg=len(signal))

    if np.sum(power) == 0:
        return 0.0

    # Computing the lower and upper limits of power bandwidth
    cum_power = np.cumsum(power)
    f_lower = freq[np.where(cum_power >= cum_power[-1] * 0.95)[0][0]]

    cum_power_inv = np.cumsum(power[::-1])
    f_upper = freq[np.abs(np.where(cum_power_inv >= cum_power[-1] * 0.95)[0][0] - len(power) + 1)]

    # Returning the bandwidth in terms of frequency

    return np.abs(f_upper - f_lower)


@set_domain("domain", "spectral")
def fft_mean_coeff(signal, fs, nfreq=256):
    """Computes the mean value of each spectrogram frequency.

    nfreq can not be higher than half signal length plus one.
    When it does, it is automatically set to half signal length plus one.

    Parameters
    ----------
    signal : nd-array
        Input from which fft mean coefficients are computed
    fs : int
        Sampling frequency
    nfreq : int
        The number of frequencies

    Returns
    -------
    nd-array
        The mean value of each spectrogram frequency

    """

    if nfreq > len(signal) // 2 + 1:
        nfreq = len(signal) // 2 + 1

    fmag_mean = scipy.signal.spectrogram(signal, fs, nperseg=nfreq * 2 - 2)[2].mean(1)

    return tuple(fmag_mean)


@set_domain("domain", "spectral")
def lpcc(signal, n_coeff=12):
    """Computes the linear prediction cepstral coefficients.

    Implementation details and description in:
    http://www.practicalcryptography.com/miscellaneous/machine-learning/tutorial-cepstrum-and-lpccs/

    Parameters
    ----------
    signal : nd-array
        Input from linear prediction cepstral coefficients are computed
    n_coeff : int
        Number of coefficients

    Returns
    -------
    nd-array
        Linear prediction cepstral coefficients

    """

    # 12-20 cepstral coefficients are sufficient for speech recognition
    lpc_coeffs = lpc(signal, n_coeff)

    if np.sum(lpc_coeffs) == 0:
        return tuple(np.zeros(n_coeff))

    # Power spectrum
    powerspectrum = np.abs(np.fft.fft(lpc_coeffs)) ** 2
    lpcc_coeff = np.fft.ifft(np.log(powerspectrum))

    return tuple(np.abs(lpcc_coeff))


@set_domain("domain", "spectral")
def spectral_entropy(signal, fs):
    """Computes the spectral entropy of the signal based on Fourier transform.

    Parameters
    ----------
    signal : nd-array
        Input from which spectral entropy is computed
    fs : int
        Sampling frequency

    Returns
    -------
    float
        The normalized spectral entropy value

    """
    # Removing DC component
    sig = signal - np.mean(signal)

    f, fmag = calc_fft(sig, fs)

    power = fmag ** 2

    if power.sum() == 0:
        return 0.0

    prob = np.divide(power, power.sum())

    prob = prob[prob != 0]

    return -np.multiply(prob, np.log2(prob)).sum() / np.log2(prob.size)


@set_domain("domain", "spectral")
def wavelet_entropy(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT entropy of the signal.

    Implementation details in:
    https://dsp.stackexchange.com/questions/13055/how-to-calculate-cwt-shannon-entropy
    B.F. Yan, A. Miyamoto, E. Bruhwiler, Wavelet transform-based modal parameter identification considering uncertainty

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    float
        wavelet entropy

    """
    if np.sum(signal) == 0:
        return 0.0

    cwt = wavelet(signal, function, widths)
    energy_scale = np.sum(np.abs(cwt), axis=1)
    t_energy = np.sum(energy_scale)
    prob = energy_scale / t_energy
    w_entropy = -np.sum(prob * np.log(prob))

    return w_entropy


@set_domain("domain", "spectral")
def wavelet_abs_mean(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT absolute mean value of each wavelet scale.

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    tuple
        CWT absolute mean value

    """

    return tuple(np.abs(np.mean(wavelet(signal, function, widths), axis=1)))


@set_domain("domain", "spectral")
def wavelet_std(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT std value of each wavelet scale.

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    tuple
        CWT std

    """

    return tuple((np.std(wavelet(signal, function, widths), axis=1)))


@set_domain("domain", "spectral")
def wavelet_var(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT variance value of each wavelet scale.

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    tuple
        CWT variance

    """

    return tuple((np.var(wavelet(signal, function, widths), axis=1)))


@set_domain("domain", "spectral")
def wavelet_energy(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT energy of each wavelet scale.

    Implementation details:
    https://stackoverflow.com/questions/37659422/energy-for-1-d-wavelet-in-python

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    tuple
        CWT energy

    """

    cwt = wavelet(signal, function, widths)
    energy = np.sqrt(np.sum(cwt ** 2, axis=1) / np.shape(cwt)[1])

    return tuple(energy)



import numpy as np
import pandas as pd
from scipy.interpolate import interp1d


def signal_window_spliter(signal, window_size, overlap):
    """Splits the signal into windows

    Parameters
    ----------
    signal : nd-array or pandas DataFrame
        input signal
    window_size :
        number of points of window size
    overlap :
        percentage of overlap, value between 0 and 1

    Returns
    -------
    list
        list of signal windows

    """

    step = int(round(window_size)) if overlap == 0 else int(round(window_size * (1 - overlap)))
    return [signal[i:i + window_size] for i in range(0, len(signal) - window_size, step)]


def merge_time_series(data, fs_resample, time_unit):
    """Time series data interpolation

    Parameters
    ----------
    data : dict
        data to interpolate
    fs_resample :
        resample sampling frequency
    time_unit :
        time unit in seconds

    Returns
    -------
    DataFrame
        Interpolated data

    """

    # time interval for interpolation
    sensors_time = np.array([[dn.iloc[0, 0], dn.iloc[-1, 0]] for k, dn in data.items()])
    t0 = np.max(sensors_time[:, 0])
    tn = np.min(sensors_time[:, 1])
    x_new = np.linspace(t0, tn, int((tn - t0) / ((1 / fs_resample) * time_unit)))

    # interpolation
    data_new = np.copy(x_new.reshape(len(x_new), 1))
    header_values = ['time']
    for k, dn in data.items():
        header_values += [k + str(i) for i in range(1, np.shape(dn)[1])]
        data_new = np.hstack((data_new, np.array([interp1d(dn.iloc[:, 0], dn.iloc[:, ax])(x_new) for ax in range(1, np.shape(dn)[1])]).T))

    return pd.DataFrame(data=data_new[:, 1:], columns=header_values[1:])


def correlation_report(features, threshold=0.95):
    """Performs a correlation report and removes highly correlated features.

    Parameters
    ----------
    features : DataFrame
        features
    threshold :
        correlation value for removing highly correlated features
    Returns
    -------
    DataFrame
        Uncorrelated features

    """
    corr_matrix = features.corr()

    inp = str(input('Do you wish to remove correlated features? Enter y/n: '))

    if inp == 'y':
        # Select upper triangle of correlation matrix
        upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(np.bool))
        # Find index and column name of features with correlation greater than 0.95
        to_drop = [column for column in upper.columns if any(upper[column] > threshold)]

        if len(to_drop) == 0:
            print('No features to remove')
        for rej in to_drop:
            print('Removing ' + str(rej))
            features = features.drop(rej, axis=1)
    return features



import json
import glob
import tsfel
import pandas as pd
from pathlib import Path


# Example of user preprocess sensor data
def pre_process(sensor_data):
    if 'Accelerometer' in sensor_data:
        sensor_data['Accelerometer'].iloc[:, 1] = sensor_data['Accelerometer'].iloc[:, 1] * 0
    return sensor_data

# DATASET DIR
main_directory = str(Path.home()) + "/Documents/Datasets/"

# JSON DIR
tsfel_path_json = tsfel.__path__[0] + '/feature_extraction/features.json'
personal_path_json = 'test_features.json'
personal_features_path = 'personal_features.py'

# DEFAULT PARAM for testing
time_unit = 1e9  # seconds
resample_rate = 30  # resample sampling frequency
window_size = 100  # number of points
overlap = 0  # varies between 0 and 1
search_criteria = ['Accelerometer.txt', 'Gyroscope.txt']
output_directory = str(Path.home()) + '/Documents/tsfel_output'
sensor_data = {}
key = 'Accelerometer'

folders = [f for f in glob.glob(main_directory + "**/", recursive=True)]
sensor_data[key] = pd.read_csv(folders[-1] + key + '.txt', header=None)

# add personal feature
tsfel.add_feature_json(personal_features_path, personal_path_json)

# Features Dictionary
settings0 = json.load(open(tsfel_path_json))
settings1 = json.load(open(personal_path_json))
settings2 = tsfel.get_features_by_domain('statistical')
settings3 = tsfel.get_features_by_domain('temporal')
settings4 = tsfel.get_features_by_domain('spectral')
settings5 = tsfel.get_features_by_domain()
settings6 = tsfel.extract_sheet('Features')
settings7 = tsfel.extract_sheet('Features_test', path_json=personal_path_json)

# Signal processing
data_new = tsfel.merge_time_series(sensor_data, resample_rate, time_unit)
windows = tsfel.signal_window_spliter(data_new, window_size, overlap)

# time_series_features_extractor
features0 = tsfel.time_series_features_extractor(settings4, windows, fs=resample_rate, window_spliter=False)
features1 = tsfel.time_series_features_extractor(settings2, data_new, fs=resample_rate, window_spliter=False, window_size=70, overlap=0.5)
features2 = tsfel.time_series_features_extractor(settings3, windows, fs=resample_rate, window_spliter=True)


# Dataset features extractor
data = tsfel.dataset_features_extractor(main_directory, settings1, search_criteria=search_criteria, time_unit=time_unit,
                                        resample_rate=resample_rate, window_size=window_size, overlap=overlap,
                                        pre_process=pre_process, output_directory=output_directory, features_path=personal_features_path)




import scipy
import numpy as np


def set_domain(key, value):
    def decorate_func(func):
        setattr(func, key, value)
        return func

    return decorate_func


def compute_time(signal, fs):
    """Creates the signal correspondent time array.

    Parameters
    ----------
    signal: nd-array
        Input from which the time is computed.
    fs: int
        Sampling Frequency

    Returns
    -------
    time : float list
        Signal time

    """

    time = range(len(signal))
    time = [float(x) / fs for x in time]
    return time


def calc_fft(signal, fs):
    """ This functions computes the fft of a signal.

    Parameters
    ----------
    signal : nd-array
        The input signal from which fft is computed
    fs : int
        Sampling frequency

    Returns
    -------
    f: nd-array
        Frequency values (xx axis)
    fmag: nd-array
        Amplitude of the frequency values (yy axis)

    """

    fmag = np.abs(np.fft.fft(signal))
    f = np.linspace(0, fs // 2, len(signal) // 2)

    return f[:len(signal) // 2].copy(), fmag[:len(signal) // 2].copy()


def filterbank(signal, fs, pre_emphasis=0.97, nfft=512, nfilt=40):
    """Computes the MEL-spaced filterbank.

    It provides the information about the power in each frequency band.

    Implementation details and description on:
    https://www.kaggle.com/ilyamich/mfcc-implementation-and-tutorial
    https://haythamfayek.com/2016/04/21/speech-processing-for-machine-learning.html#fnref:1

    Parameters
    ----------
    signal : nd-array
        Input from which filterbank is computed
    fs : int
        Sampling frequency
    pre_emphasis : float
        Pre-emphasis coefficient for pre-emphasis filter application
    nfft : int
        Number of points of fft
    nfilt : int
        Number of filters

    Returns
    -------
    nd-array
        MEL-spaced filterbank

    """

    # Signal is already a window from the original signal, so no frame is needed.
    # According to the references it is needed the application of a window function such as
    # hann window. However if the signal windows don't have overlap, we will lose information,
    # as the application of a hann window will overshadow the windows signal edges.

    # pre-emphasis filter to amplify the high frequencies

    emphasized_signal = np.append(np.array(signal)[0], np.array(signal[1:]) - pre_emphasis * np.array(signal[:-1]))

    # Fourier transform and Power spectrum
    mag_frames = np.absolute(np.fft.rfft(emphasized_signal, nfft))  # Magnitude of the FFT

    pow_frames = ((1.0 / nfft) * (mag_frames ** 2))  # Power Spectrum

    low_freq_mel = 0
    high_freq_mel = (2595 * np.log10(1 + (fs / 2) / 700))  # Convert Hz to Mel
    mel_points = np.linspace(low_freq_mel, high_freq_mel, nfilt + 2)  # Equally spaced in Mel scale
    hz_points = (700 * (10 ** (mel_points / 2595) - 1))  # Convert Mel to Hz
    filter_bin = np.floor((nfft + 1) * hz_points / fs)

    fbank = np.zeros((nfilt, int(np.floor(nfft / 2 + 1))))
    for m in range(1, nfilt + 1):

        f_m_minus = int(filter_bin[m - 1])  # left
        f_m = int(filter_bin[m])  # center
        f_m_plus = int(filter_bin[m + 1])  # right

        for k in range(f_m_minus, f_m):
            fbank[m - 1, k] = (k - filter_bin[m - 1]) / (filter_bin[m] - filter_bin[m - 1])
        for k in range(f_m, f_m_plus):
            fbank[m - 1, k] = (filter_bin[m + 1] - k) / (filter_bin[m + 1] - filter_bin[m])

    # Area Normalization
    # If we don't normalize the noise will increase with frequency because of the filter width.
    enorm = 2.0 / (hz_points[2:nfilt + 2] - hz_points[:nfilt])
    fbank *= enorm[:, np.newaxis]

    filter_banks = np.dot(pow_frames, fbank.T)
    filter_banks = np.where(filter_banks == 0, np.finfo(float).eps, filter_banks)  # Numerical Stability
    filter_banks = 20 * np.log10(filter_banks)  # dB

    return filter_banks


def autocorr_norm(signal):
    """Computes the autocorrelation.

    Implementation details and description in:
    https://ccrma.stanford.edu/~orchi/Documents/speaker_recognition_report.pdf

    Parameters
    ----------
    signal : nd-array
        Input from linear prediction coefficients are computed

    Returns
    -------
    nd-array
        Autocorrelation result

    """

    signal = signal - np.mean(signal)
    r = np.correlate(signal, signal, mode='full')[-len(signal):]

    if np.sum(signal) == 0:
        return np.zeros(len(signal))

    acf = r / (np.var(signal) * (np.arange(len(signal), 0, -1)))

    return acf


def create_symmetric_matrix(acf, n_coeff=12):
    """Computes a symmetric matrix.

    Implementation details and description in:
    https://ccrma.stanford.edu/~orchi/Documents/speaker_recognition_report.pdf

    Parameters
    ----------
    acf : nd-array
        Input from which a symmetric matrix is computed
    n_coeff : int
        Number of coefficients

    Returns
    -------
    nd-array
        Symmetric Matrix

    """

    smatrix = np.empty((n_coeff, n_coeff))

    for i in range(n_coeff):
        for j in range(n_coeff):
            smatrix[i, j] = acf[np.abs(i - j)]
    return smatrix


def lpc(signal, n_coeff=12):
    """Computes the linear prediction coefficients.

    Implementation details and description in:
    https://ccrma.stanford.edu/~orchi/Documents/speaker_recognition_report.pdf

    Parameters
    ----------
    signal : nd-array
        Input from linear prediction coefficients are computed
    n_coeff : int
        Number of coefficients

    Returns
    -------
    nd-array
        Linear prediction coefficients

    """

    # Calculate LPC with Yule-Walker
    acf = autocorr_norm(signal)
    r = -acf[1:n_coeff + 1].T
    smatrix = create_symmetric_matrix(acf, n_coeff)
    if np.sum(smatrix) == 0:
        return tuple(np.zeros(n_coeff))

    lpc_coeffs = np.dot(np.linalg.inv(smatrix), r)
    lpc_coeffs = lpc_coeffs / np.max(np.abs(lpc_coeffs))

    return tuple(lpc_coeffs)


def create_xx(features):
    """Computes the range of features amplitude for the probability density function calculus.

    Parameters
    ----------
    features : nd-array
        Input features

    Returns
    -------
    nd-array
        range of features amplitude

    """

    features_ = np.copy(features)

    if max(features_) < 0:
        max_f = - max(features_)
        min_f = min(features_)
    else:
        min_f = min(features_)
        max_f = max(features_)

    if min(features_) == max(features_):
        xx = np.linspace(min_f, min_f + 10, len(features_))
    else:
        xx = np.linspace(min_f, max_f, len(features_))

    return xx


def kde(features):
    """Computes the probability density function of the input signal using a Gaussian KDE (Kernel Density Estimate)

    Parameters
    ----------
    features : nd-array
        Input from which probability density function is computed

    Returns
    -------
    nd-array
        probability density values

    """
    features_ = np.copy(features)
    xx = create_xx(features_)

    if min(features_) == max(features_):
        noise = np.random.randn(len(features_)) * 0.0001
        features_ = np.copy(features_ + noise)

    kernel = scipy.stats.gaussian_kde(features_, bw_method='silverman')

    return np.array(kernel(xx) / np.sum(kernel(xx)))


def gaussian(features):
    """Computes the probability density function of the input signal using a Gaussian function

    Parameters
    ----------
    features : nd-array
        Input from which probability density function is computed
    Returns
    -------
    nd-array
        probability density values

    """

    features_ = np.copy(features)

    xx = create_xx(features_)
    std_value = np.std(features_)
    mean_value = np.mean(features_)

    if std_value == 0:
        return 0.0
    pdf_gauss = scipy.stats.norm.pdf(xx, mean_value, std_value)

    return np.array(pdf_gauss / np.sum(pdf_gauss))


def wavelet(signal, function=scipy.signal.ricker, widths=np.arange(1, 10)):
    """Computes CWT (continuous wavelet transform) of the signal.

    Parameters
    ----------
    signal : nd-array
        Input from which CWT is computed
    function :  wavelet function
        Default: scipy.signal.ricker
    widths :  nd-array
        Widths to use for transformation
        Default: np.arange(1,10)

    Returns
    -------
    nd-array
        The result of the CWT along the time axis
        matrix with size (len(widths),len(signal))

    """

    if isinstance(function, str):
        function = eval(function)

    if isinstance(widths, str):
        widths = eval(widths)

    cwt = scipy.signal.cwt(signal, function, widths)

    return cwt


def calc_ecdf(signal):
    """Computes the ECDF of the signal.

      Parameters
      ----------
      signal : nd-array
          Input from which ECDF is computed
      Returns
      -------
      nd-array
        Sorted signal and computed ECDF.

      """
    return np.sort(signal), np.arange(1, len(signal)+1)/len(signal)




import time
import json
import numpy as np
from scipy.optimize import curve_fit
from tsfel.feature_extraction.features_settings import load_json
from tsfel.feature_extraction.calc_features import calc_window_features


# curves
def n_squared(x, no):
    """The model function"""
    return no * x ** 2


def n_nlog(x, no):
    """The model function"""
    return no * x * np.log(x)


def n_linear(x, no):
    """The model function"""
    return no * x


def n_log(x, no):
    """The model function"""
    return no * np.log(x)


def n_constant(x, no):
    """The model function"""
    return np.zeros(len(x)) + no


def find_best_curve(t, signal):
    """Finds the best curve.

    Parameters
    ----------
    t : nd-array
        Log space
    signal : nd-array
        Mean execution time array

    Returns
    -------
    str
        Best fit curve name

    """

    all_chisq = []
    list_curves = [n_squared, n_nlog, n_linear, n_log, n_constant]
    all_curves = []
    # Model parameters
    stdev = 2
    sig = np.zeros(len(signal)) + stdev

    # Fit the curve
    for curve in list_curves:
        start = 1
        popt, pcov = curve_fit(curve, t, signal, sigma=sig, p0=start, absolute_sigma=True)

        # Compute chi square
        nexp = curve(t, *popt)
        r = signal - nexp
        chisq = np.sum((r / stdev) ** 2)
        all_chisq.append(chisq)
        all_curves.append(nexp)

    idx_best = np.argmin(all_chisq)

    curve_name = str(list_curves[idx_best])
    idx1 = curve_name.find("n_")
    idx2 = curve_name.find("at")
    curve_name = curve_name[idx1 + 2:idx2 - 1]

    return curve_name


def compute_complexity(feature, domain, json_path, **kwargs):
    """Computes the feature complexity.

    Parameters
    ----------
    feature : string
        Feature name
    domain : string
        Feature domain
    json_path: json
        Features json file
    \**kwargs:
    See below:
        * *features_path* (``string``) --
            Directory of script with personal features

    Returns
    -------
    int
        Feature complexity

    Writes complexity in json file

    """

    dictionary = load_json(json_path)

    features_path = kwargs.get('features_path', None)

    # The inputs from this function should be replaced by a dictionary
    one_feat_dict = {domain: {feature: dictionary[domain][feature]}}

    t = np.logspace(3.0, 5.0, 6)
    signal, s = [], []
    f = 0.05
    x = np.arange(0, t[-1] + 1, 1)
    fs = 100
    wave = np.sin(2 * np.pi * f * x / fs)

    for ti in t:
        for _ in range(20):

            start = time.time()
            calc_window_features(one_feat_dict, wave[:int(ti)], fs, features_path=features_path)
            end = time.time()

            s += [end - start]

        signal += [np.mean(s)]

    curve_name = find_best_curve(t, signal)
    dictionary[domain][feature]['complexity'] = curve_name

    with open(json_path, "w") as write_file:
        json.dump(dictionary, write_file, indent=4, sort_keys=True)

    if curve_name == 'constant' or curve_name == 'log':
        return 1
    elif curve_name == 'linear':
        return 2
    elif curve_name == 'nlog' or curve_name == 'squared':
        return 3
    else:
        return 0



from tsfel.feature_extraction.calc_features import *
from tsfel.feature_extraction.features import *
from tsfel.feature_extraction.features_settings import *
from tsfel.feature_extraction.features_utils import *



import json
import tsfel


def load_json(json_path):
    """Loads the json file given by filename.

    Parameters
    ----------
    json_path : string
        Json path

    Returns
    -------
    Dict
        Dictionary

    """

    return json.load(open(json_path))


def get_features_by_domain(domain=None, json_path=None):
    """Creates a dictionary with the features settings by domain.

    Parameters
    ----------
    domain : string
        Available domains: "statistical"; "spectral"; "temporal"
        If domain equals None, then the features settings from all domains are returned.
    json_path : string
        Directory of json file. Default: package features.json directory

    Returns
    -------
    Dict
        Dictionary with the features settings

    """

    if json_path is None:
        json_path = tsfel.__path__[0]+"/feature_extraction/features.json"

        if domain not in ['statistical', 'temporal', 'spectral', None]:
            raise SystemExit(
                'No valid domain. Choose: statistical, temporal, spectral or None (for all feature settings).')

    dict_features = load_json(json_path)
    if domain is None:
        return dict_features
    else:
        return {domain: dict_features[domain]}






import ast
import tsfel
import gspread
import numpy as np
from oauth2client.service_account import ServiceAccountCredentials
from tsfel.feature_extraction.features_settings import load_json
from tsfel.utils.calculate_complexity import compute_complexity


def filter_features(features, filters):
    """Filtering features based on Google sheet.

    Parameters
    ----------
    features : dict
        Dictionary with features
    filters : dict
        Filters from Google sheets

    Returns
    -------
    dict
        Filtered features

    """
    features_all = list(np.concatenate([list(features[dk].keys()) for dk in sorted(features.keys())]))
    list_shown, feat_shown = list(features.keys()), features_all
    cost_shown = features_all
    if filters['2'] != {}:
        list_hidden = filters['2']['hiddenValues']
        list_shown = [dk for dk in features.keys() if dk not in list_hidden]
    if filters['1'] != {}:
        feat_hidden = filters['1']['hiddenValues']
        feat_shown = [ff for ff in features_all if ff not in feat_hidden]
    if filters['3'] != {}:
        cost_numbers = filters['3']['hiddenValues']
        cost_hidden = list(np.concatenate([['constant', 'log'] if int(cn) == 1 else
                                           ['squared', 'nlog'] if int(cn) == 3 else ['linear']
                                           if int(cn) == 2 else ['unknown'] for cn in cost_numbers]))
        cost_shown = []
        for dk in features.keys():
            cost_shown += [ff for ff in features[dk].keys() if features[dk][ff]['complexity'] not in cost_hidden]
    features_filtered = list(np.concatenate([list(features[dk].keys())
                                             for dk in sorted(features.keys()) if dk in list_shown]))
    features_filtered = [ff for ff in features_filtered if ff in feat_shown]
    features_filtered = [cc for cc in features_filtered if cc in cost_shown]

    return features_filtered


def extract_sheet(gsheet_name, **kwargs):
    """Interaction between features.json and Google sheets.

    Parameters
    ----------
    gsheet_name : str
        Google Sheet name
    \**kwargs:
    See below:
        * *path_json* (``string``) --
            Json path
    Returns
    -------
    dict
        Features

    """
    # Path to Tsfel
    lib_path = tsfel.__path__

    # Access features.json
    path_json = kwargs.get('path_json', lib_path[0] + '/feature_extraction/features.json')

    # Read features.json into a dictionary of features and parameters
    dict_features = load_json(path_json)

    # Number of features from json file
    len_json = 0
    for domain in list(dict_features.keys()):
        len_json += len(dict_features[domain].keys())

    # Access Google sheet
    # Scope and credentials using the content of client_secret.json file
    scope = ['https://spreadsheets.google.com/feeds',
             'https://www.googleapis.com/auth/drive']
    creds = ServiceAccountCredentials.from_json_keyfile_name(lib_path[0] + '/utils/client_secret.json', scope)

    # Create a gspread client authorizing it using those credentials
    client = gspread.authorize(creds)

    # and pass it to the spreadsheet name, getting access to sheet1
    confManager = client.open(gsheet_name)
    sheet = confManager.sheet1
    metadata = confManager.fetch_sheet_metadata()

    # Reading from Google Sheet
    # Features
    list_of_features = sheet.col_values(2)[4:]

    try:
        filters = metadata['sheets'][sheet.id]['basicFilter']['criteria']
        list_filt_features = filter_features(dict_features, filters)
    except KeyError:
        print('No filters running. Check Google Sheet filters.')
        list_filt_features = list_of_features.copy()

    use_or_not = ['TRUE' if lf in list_filt_features else 'FALSE' for lf in list_of_features]

    assert len(list_of_features) <= (len_json), \
        "To insert a new feature, please add it to data/features.json with the code in src/utils/features.py"

    # adds a new feature in Google sheet if it is missing from features.json
    if len(list_of_features) < (len_json):

        # new feature was added
        for domain in dict_features.keys():
            for feat in dict_features[domain].keys():
                if feat not in list_of_features:
                    feat_dict = dict_features[domain][feat]
                    param = ''
                    fs = 'no'

                    # Read parameters from features.json
                    if feat_dict['parameters']:
                        param = feat_dict['parameters'].copy()
                        if 'fs' in feat_dict['parameters']:
                            fs = 'yes'
                            param.pop('fs')
                            if len(param) == 0:
                                param = ''

                    curve = feat_dict['complexity']
                    curves_all = ['linear', 'log', 'square', 'nlog', 'constant']
                    complexity = compute_complexity(feat, domain,
                                                    path_json) if curve not in curves_all else 1 if curve in [
                        'constant', 'log'] else 2 if curve == 'linear' else 3
                    new_feat = ['', feat, domain, complexity, fs, str(param),
                                feat_dict['description']]

                    # checks if the Google sheet has no features
                    if sheet.findall(domain) == []:
                        idx_row = 4
                    else:
                        idx_row = sheet.findall(domain)[-1].row

                    # Add new feature at the end of feature domain
                    sheet.insert_row(new_feat, idx_row + 1)
                    print(feat + " feature was added to Google Sheet.")

        # Update list of features and domains from Google sheet
        list_of_features = sheet.col_values(2)[4:]

        # Update filtered features from Google sheet. Check if filters exist.
        try:
            filters = metadata['sheets'][sheet.id]['basicFilter']['criteria']
            list_filt_features = filter_features(dict_features, filters)
        except KeyError:
            list_filt_features = list_of_features.copy()

        use_or_not = ['TRUE' if lf in list_filt_features else 'FALSE' for lf in list_of_features]

    assert 'TRUE' in use_or_not, 'Please select a feature to extract!' + '\n'

    # Reading from Google Sheet
    # Domain
    list_domain = sheet.col_values(3)[4:]
    # Parameters and fs
    gs_param_list = sheet.col_values(6)[4:]
    gs_fs_list = sheet.col_values(5)[4:]
    gs_fs = int(sheet.cell(4, 9).value)

    # Fix for empty cells in parameters column
    if len(gs_param_list) < len(list_of_features):
        empty = [''] * (len(list_of_features) - len(gs_param_list))
        gs_param_list = gs_param_list + empty

    # Update dict of features with changes from Google sheet
    for ii, feature in enumerate(list_of_features):
        domain = list_domain[ii]
        try:
            if use_or_not[ii] == 'TRUE':
                dict_features[domain][feature]['use'] = 'yes'
                # Check features parameters from Google sheet
                if gs_param_list[ii] != '':
                    param_sheet = ast.literal_eval(gs_param_list[ii])
                    # update dic of features based on Google sheet
                    dict_features[domain][feature]['parameters'] = param_sheet
                # Check features that use sampling frequency parameter
                if gs_fs_list[ii] != 'no':
                    # update dict of features based on Google sheet fs
                    dict_features[domain][feature]['parameters']['fs'] = gs_fs
            else:
                dict_features[domain][feature]['use'] = 'no'
        except KeyError:
            print('Unknown domain at cell', int(ii + 5))

    return dict_features



import os
import sys
import glob
import numbers
import pathlib
import warnings
import importlib
import numpy as np
import pandas as pd
from pathlib import Path
from tsfel.utils.signal_processing import merge_time_series, signal_window_spliter


def dataset_features_extractor(main_directory, feat_dict, verbose=1, **kwargs):
    """Extracts features from a dataset.

    Parameters
    ----------
    main_directory : String
        Input directory
    feat_dict : dict
        Dictionary with features
    verbose : int
        Level of function communication
        (0 or 1 (Default))
    \**kwargs:
    See below:
        * *search_criteria* (``list``) --
            List of file names to compute features. (Example: 'Accelerometer.txt')
            (default: ``None``)

        * *time_unit* (``float``) --
            Time unit
            (default: ``1e9``)

        * *resampling_rate* (``int``) --
            Resampling rate
            (default: ``100``)

        * *window_size* (``int``) --
            Window size in number of samples
            (default: ``100``)

        * *overlap* (``float``) --
            Overlap between 0 and 1
            (default: ``0``)

        * *pre_process* (``function``) --
            Function with pre processing code

            (default: ``None``)

        * *output_directory* (``String``) --
            Output directory
            (default: ``'output_directory', str(Path.home()) + '/tsfel_output'``)

    Returns
    -------
    file
        csv file with the extracted features

    """
    search_criteria = kwargs.get('search_criteria', None)
    time_unit = kwargs.get('time_unit', 1e9)
    resample_rate = kwargs.get('resample_rate', 30)
    window_size = kwargs.get('window_size', 100)
    overlap = kwargs.get('overlap', 0)
    pre_process = kwargs.get('pre_process', None)
    output_directory = kwargs.get('output_directory', str(Path.home()) + '/tsfel_output')
    features_path = kwargs.get('features_path', None)

    if main_directory[-1] != os.sep:
        main_directory = main_directory+os.sep

    folders = [f for f in glob.glob(main_directory + "**/", recursive=True)]

    for fl in folders:
        sensor_data = {}
        if search_criteria:
            for c in search_criteria:
                if os.path.isfile(fl + c):
                    key = c.split('.')[0]
                    sensor_data[key] = pd.read_csv(fl + c, header=None)
        else:
            all_files = np.concatenate((glob.glob(fl + '/*.txt'), glob.glob(fl + '/*.csv')))
            for c in all_files:
                key = c.split(os.sep)[-1].split('.')[0]
                try:
                    data_file = pd.read_csv(c, header=None)
                except pd.io.common.CParserError:
                    continue

                if np.dtype('O') in np.array(data_file.dtypes):
                    continue

                sensor_data[key] = pd.read_csv(c, header=None)

        if not sensor_data:
            continue

        pp_sensor_data = sensor_data if pre_process is None else pre_process(sensor_data)

        data_new = merge_time_series(pp_sensor_data, resample_rate, time_unit)

        windows = signal_window_spliter(data_new, window_size, overlap)

        if features_path:
            features = time_series_features_extractor(feat_dict, windows, fs=resample_rate, verbose=0, features_path=features_path)
        else:
            features = time_series_features_extractor(feat_dict, windows, fs=resample_rate, verbose=0)

        pathlib.Path(output_directory + fl).mkdir(parents=True, exist_ok=True)
        features.to_csv(output_directory + fl + '/Features.csv', sep=',', encoding='utf-8')

    if verbose == 1:
        print('Features files saved in: ', output_directory)


def time_series_features_extractor(dict_features, signal_windows, fs=None, window_spliter=False, verbose=1, **kwargs):
    """Extraction of time series features.

    Parameters
    ----------
    dict_features : dict
        Dictionary with features
    signal_windows: list
        Input from which features are computed, window
    fs : int or None
        Sampling frequency
    window_spliter: bool
        If True computes the signal windows
    verbose : int
        Level of function communication
        (0 or 1 (Default))
    \**kwargs:
    See below:
        * *window_size* (``int``) --
            Window size in number of samples
            (default: ``100``)

        * *overlap* (``float``) --
            Overlap between 0 and 1
            (default: ``0``)

        * *features_path* (``string``) --
            Directory of script with personal features

    Returns
    -------
    DataFrame
        Extracted features

    """

    if verbose == 1:
        print("*** Feature extraction started ***")

    window_size = kwargs.get('window_size', 100)
    overlap = kwargs.get('overlap', 0)
    features_path = kwargs.get('features_path', None)

    feat_val = pd.DataFrame()
    if window_spliter:
        signal_windows = signal_window_spliter(signal_windows, window_size, overlap)

    if isinstance(signal_windows, pd.DataFrame):
        features = calc_window_features(dict_features, signal_windows, fs, features_path=features_path)
        feat_val = feat_val.append(features)
    else:
        for wind_sig in signal_windows:
            if isinstance(wind_sig, numbers.Real):
                features = calc_window_features(dict_features, signal_windows, fs, features_path=features_path)
                feat_val = feat_val.append(features)
                break
            else:
                features = calc_window_features(dict_features, wind_sig, fs, features_path=features_path)
                feat_val = feat_val.append(features)

    if verbose == 1:
        print("*** Feature extraction finished ***")

    return feat_val.reset_index(drop=True)


def calc_window_features(dict_features, signal_window, fs, **kwargs):
    """This function computes features matrix for one window.

    Parameters
    ----------
    dict_features : dict
        Dictionary with features
    signal_window: pandas DataFrame
        Input from which features are computed, window
    fs : int
        Sampling frequency
    \**kwargs:
    See below:
        * *features_path* (``string``) --
            Directory of script with personal features

    Returns
    -------
    pandas DataFrame
        (columns) names of the features
        (data) values of each features for signal

    """

    features_path = kwargs.get('features_path', None)

    # Execute imports
    exec("import tsfel")
    domain = dict_features.keys()

    if features_path:
        sys.path.append(features_path[:-len(features_path.split(os.sep)[-1])-1])
        exec("import "+features_path.split(os.sep)[-1][:-3])
        importlib.reload(sys.modules[features_path.split(os.sep)[-1][:-3]])
        exec("from " + features_path.split(os.sep)[-1][:-3]+" import *")

    # Create global arrays
    func_total = []
    func_names = []
    imports_total = []
    parameters_total = []
    feature_results = []
    feature_names = []

    for _type in domain:
        domain_feats = dict_features[_type].keys()

        for feat in domain_feats:

            # Only returns used functions
            if dict_features[_type][feat]['use'] == 'yes':

                # Read Function Name (generic name)
                func_names = [feat]

                # Read Function (real name of function)
                func_total = [dict_features[_type][feat]['function']]

                # Check for parameters
                if dict_features[_type][feat]['parameters'] != '':
                    param = dict_features[_type][feat]['parameters']

                    # Check assert fs parameter:
                    if 'fs' in param:

                        # Select which fs to use
                        if fs is None:

                            # Check if features dict has default sampling frequency value
                            if type(param['fs']) is int or type(param['fs']) is float:
                                parameters_total = [str(key) + '=' + str(value) for key, value in param.items()]

                                # raise a warning
                                warnings.warn('Using default sampling frequency: '+str(param['fs'])+" Hz.", stacklevel=2)
                            else:
                                raise SystemExit('No sampling frequency assigned.')
                        else:
                            parameters_total = [str(key) + '=' + str(value) for key, value in param.items()
                                                if key not in 'fs']
                            parameters_total += ['fs =' + str(fs)]

                    # feature has no fs parameter
                    else:
                        parameters_total = []
                        for key, value in param.items():
                            if type(value) is str:
                                value = '"'+value+'"'
                            parameters_total.append([str(key) + '=' + str(value)])
                else:
                    parameters_total = ''

                # Name of each column to be concatenate with feature name
                if not isinstance(signal_window, pd.DataFrame):
                    signal_window = pd.DataFrame(data=signal_window)
                header_names = signal_window.columns.values

                for ax in range(len(header_names)):
                    window = signal_window.iloc[:, ax]
                    execf = func_total[0] + '(window'

                    if parameters_total != '':
                        execf += ', ' + str(parameters_total).translate(str.maketrans({'[': '', ']': '', "'": ''}))

                    execf += ')'
                    eval_result = eval(execf, locals())

                    # Function returns more than one element
                    if type(eval_result) == tuple:
                        for rr in range(len(eval_result)):
                            if np.isnan(eval_result[0]):
                                eval_result = np.zeros(len(eval_result))
                            feature_results += [eval_result[rr]]
                            feature_names += [str(header_names[ax]) + '_' + func_names[0] + '_' + str(rr)]
                    else:
                        feature_results += [eval_result]
                        feature_names += [str(header_names[ax]) + '_' + func_names[0]]

    features = pd.DataFrame(data=np.array(feature_results).reshape(1, len(feature_results)),
                            columns=np.array(feature_names))

    return features



from tsfel.feature_extraction.features import *
from numpy.testing import run_module_suite

from tsfel.feature_extraction.features import *

# Implementing signals for testing features

const0 = np.zeros(20)
const1 = np.ones(20)
constNeg = np.ones(20) * (-1)
constF = np.ones(20) * 2.5
lin = np.arange(20)
lin0 = np.linspace(-10, 10, 20)
f = 5
sample = 1000
x = np.arange(0, sample, 1)
Fs = 1000
wave = np.sin(2 * np.pi * f * x / Fs)
np.random.seed(seed=10)
noiseWave = wave + np.random.normal(0, 0.1, 1000)
offsetWave = wave + 2


# ############################################### STATISTICAL FEATURES ############################################### #
def test_hist():
    np.testing.assert_almost_equal(hist(const0, 10, 5), (0.0, 0.0, 0.0, 0.0, 0.0, 20.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(hist(const1, 10, 5), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 20.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(hist(constNeg, 10, 5), (0.0, 0.0, 0.0, 0.0, 20.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(hist(constF, 10, 5), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 20.0, 0.0, 0.0))
    np.testing.assert_almost_equal(hist(lin, 10, 5), (0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 2))
    np.testing.assert_almost_equal(hist(wave, 10, 5), (0.0, 0.0, 0.0, 0.0, 499, 496, 5, 0.0, 0.0, 0.0), decimal=5)
    np.testing.assert_almost_equal(hist(offsetWave, 10, 5), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 499, 496, 5, 0.0), decimal=5)
    np.testing.assert_almost_equal(hist(noiseWave, 10, 5), (0.0, 0.0, 0.0, 48, 446, 450, 56, 0.0, 0.0, 0.0), decimal=5)


def test_skewness():
    np.testing.assert_almost_equal(skewness(const0), 0.0)
    np.testing.assert_almost_equal(skewness(const1), 0.0)
    np.testing.assert_almost_equal(skewness(constNeg), 0.0)
    np.testing.assert_almost_equal(skewness(constF), 0.0)
    np.testing.assert_almost_equal(skewness(lin), 0)
    np.testing.assert_almost_equal(skewness(lin0), -1.0167718723297815e-16, decimal=5)
    np.testing.assert_almost_equal(skewness(wave), -2.009718347115232e-17, decimal=5)
    np.testing.assert_almost_equal(skewness(offsetWave), 9.043732562018544e-16, decimal=5)
    np.testing.assert_almost_equal(skewness(noiseWave), -0.0004854111290521465, decimal=5)


def test_kurtosis():
    np.testing.assert_almost_equal(kurtosis(const0), -3)
    np.testing.assert_almost_equal(kurtosis(const1), -3)
    np.testing.assert_almost_equal(kurtosis(constNeg), -3)
    np.testing.assert_almost_equal(kurtosis(constF), -3.0)
    np.testing.assert_almost_equal(kurtosis(lin), -1.206015037593985, decimal=2)
    np.testing.assert_almost_equal(kurtosis(lin0), -1.2060150375939847, decimal=2)
    np.testing.assert_almost_equal(kurtosis(wave), -1.501494077162359, decimal=2)
    np.testing.assert_almost_equal(kurtosis(offsetWave), -1.5014940771623597, decimal=2)
    np.testing.assert_almost_equal(kurtosis(noiseWave), -1.4606204906023366, decimal=2)


def test_mean():
    np.testing.assert_almost_equal(calc_mean(const0), 0.0)
    np.testing.assert_almost_equal(calc_mean(const1), 1.0)
    np.testing.assert_almost_equal(calc_mean(constNeg), -1.0)
    np.testing.assert_almost_equal(calc_mean(constF), 2.5)
    np.testing.assert_almost_equal(calc_mean(lin), 9.5)
    np.testing.assert_almost_equal(calc_mean(lin0), -3.552713678800501e-16, decimal=5)
    np.testing.assert_almost_equal(calc_mean(wave), 7.105427357601002e-18, decimal=5)
    np.testing.assert_almost_equal(calc_mean(offsetWave), 2.0, decimal=5)
    np.testing.assert_almost_equal(calc_mean(noiseWave), -0.0014556635615470554, decimal=5)


def test_median():
    np.testing.assert_almost_equal(calc_median(const0), 0.0)
    np.testing.assert_almost_equal(calc_median(const1), 1.0)
    np.testing.assert_almost_equal(calc_median(constNeg), -1.0)
    np.testing.assert_almost_equal(calc_median(constF), 2.5)
    np.testing.assert_almost_equal(calc_median(lin), 9.5)
    np.testing.assert_almost_equal(calc_median(lin0), -3.552713678800501e-16, decimal=5)
    np.testing.assert_almost_equal(calc_median(wave), 7.105427357601002e-18, decimal=5)
    np.testing.assert_almost_equal(calc_median(offsetWave), 2.0, decimal=5)
    np.testing.assert_almost_equal(calc_median(noiseWave), 0.013846093997438328, decimal=5)


def test_max():
    np.testing.assert_almost_equal(calc_max(const0), 0.0)
    np.testing.assert_almost_equal(calc_max(const1), 1.0)
    np.testing.assert_almost_equal(calc_max(constNeg), -1.0)
    np.testing.assert_almost_equal(calc_max(constF), 2.5)
    np.testing.assert_almost_equal(calc_max(lin), 19)
    np.testing.assert_almost_equal(calc_max(lin0), 10.0, decimal=5)
    np.testing.assert_almost_equal(calc_max(wave), 1.0, decimal=5)
    np.testing.assert_almost_equal(calc_max(noiseWave), 1.221757617217142, decimal=5)
    np.testing.assert_almost_equal(calc_max(offsetWave), 3.0, decimal=5)


def test_min():
    np.testing.assert_almost_equal(calc_min(const0), 0.0)
    np.testing.assert_almost_equal(calc_min(const1), 1.0)
    np.testing.assert_almost_equal(calc_min(constNeg), -1.0)
    np.testing.assert_almost_equal(calc_min(constF), 2.5)
    np.testing.assert_almost_equal(calc_min(lin), 0)
    np.testing.assert_almost_equal(calc_min(lin0), -10.0, decimal=5)
    np.testing.assert_almost_equal(calc_min(wave), -1.0, decimal=5)
    np.testing.assert_almost_equal(calc_min(noiseWave), -1.2582533627830566, decimal=5)
    np.testing.assert_almost_equal(calc_min(offsetWave), 1.0, decimal=5)


def test_variance():
    np.testing.assert_almost_equal(calc_var(const0), 0.0)
    np.testing.assert_almost_equal(calc_var(const1), 0.0)
    np.testing.assert_almost_equal(calc_var(constNeg), 0.0)
    np.testing.assert_almost_equal(calc_var(constF), 0.0)
    np.testing.assert_almost_equal(calc_var(lin), 33.25)
    np.testing.assert_almost_equal(calc_var(lin0), 36.84210526315789, decimal=5)
    np.testing.assert_almost_equal(calc_var(wave), 0.5, decimal=5)
    np.testing.assert_almost_equal(calc_var(offsetWave), 0.5, decimal=5)
    np.testing.assert_almost_equal(calc_var(noiseWave), 0.5081167177369529, decimal=5)


def test_std():
    np.testing.assert_almost_equal(calc_std(const0), 0.0)
    np.testing.assert_almost_equal(calc_std(const1), 0.0)
    np.testing.assert_almost_equal(calc_std(constNeg), 0.0)
    np.testing.assert_almost_equal(calc_std(constF), 0.0)
    np.testing.assert_almost_equal(calc_std(lin), 5.766281297335398)
    np.testing.assert_almost_equal(calc_std(lin0), 6.069769786668839, decimal=5)
    np.testing.assert_almost_equal(calc_std(wave), 0.7071067811865476, decimal=5)
    np.testing.assert_almost_equal(calc_std(offsetWave), 0.7071067811865476, decimal=5)
    np.testing.assert_almost_equal(calc_std(noiseWave), 0.7128230620125536, decimal=5)


def test_interq_range():
    np.testing.assert_almost_equal(interq_range(const0), 0.0)
    np.testing.assert_almost_equal(interq_range(const1), 0.0)
    np.testing.assert_almost_equal(interq_range(constNeg), 0.0)
    np.testing.assert_almost_equal(interq_range(constF), 0.0)
    np.testing.assert_almost_equal(interq_range(lin), 9.5)
    np.testing.assert_almost_equal(interq_range(lin0), 10.0, decimal=5)
    np.testing.assert_almost_equal(interq_range(wave), 1.414213562373095, decimal=5)
    np.testing.assert_almost_equal(interq_range(offsetWave), 1.414213562373095, decimal=5)
    np.testing.assert_almost_equal(interq_range(noiseWave), 1.4277110228590328, decimal=5)


def test_mean_abs_diff():
    np.testing.assert_almost_equal(mean_abs_diff(const0), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(const1), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(constF), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(lin), 1.0)
    np.testing.assert_almost_equal(mean_abs_diff(lin0), 1.0526315789473684, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(wave), 0.019988577818740614, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(noiseWave), 0.10700252903161511, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(offsetWave), 0.019988577818740614, decimal=5)


def test_mean_abs_deviation():
    np.testing.assert_almost_equal(mean_abs_deviation(const0), 0.0)
    np.testing.assert_almost_equal(mean_abs_deviation(const1), 0.0)
    np.testing.assert_almost_equal(mean_abs_deviation(constNeg), 0.0)
    np.testing.assert_almost_equal(mean_abs_deviation(constF), 0.0)
    np.testing.assert_almost_equal(mean_abs_deviation(lin), 5.0)
    np.testing.assert_almost_equal(mean_abs_deviation(lin0), 5.263157894736842, decimal=5)
    np.testing.assert_almost_equal(mean_abs_deviation(wave), 0.6365674116287157, decimal=5)
    np.testing.assert_almost_equal(mean_abs_deviation(noiseWave), 0.6392749078483896, decimal=5)
    np.testing.assert_almost_equal(mean_abs_deviation(offsetWave), 0.6365674116287157, decimal=5)


def test_calc_median_abs_deviation():
    np.testing.assert_almost_equal(median_abs_deviation(const0), 0.0)
    np.testing.assert_almost_equal(median_abs_deviation(const1), 0.0)
    np.testing.assert_almost_equal(median_abs_deviation(constNeg), 0.0)
    np.testing.assert_almost_equal(median_abs_deviation(constF), 0.0)
    np.testing.assert_almost_equal(median_abs_deviation(lin), 5.0)
    np.testing.assert_almost_equal(median_abs_deviation(lin0), 5.2631578947368425, decimal=5)
    np.testing.assert_almost_equal(median_abs_deviation(wave), 0.7071067811865475, decimal=5)
    np.testing.assert_almost_equal(median_abs_deviation(offsetWave), 0.7071067811865475, decimal=5)
    np.testing.assert_almost_equal(median_abs_deviation(noiseWave), 0.7068117164205888, decimal=5)


def test_rms():
    np.testing.assert_almost_equal(rms(const0), 0.0)
    np.testing.assert_almost_equal(rms(const1), 1.0)
    np.testing.assert_almost_equal(rms(constNeg), 1.0)
    np.testing.assert_almost_equal(rms(constF), 2.5)
    np.testing.assert_almost_equal(rms(lin), 11.113055385446435)
    np.testing.assert_almost_equal(rms(lin0), 6.06976978666884, decimal=5)
    np.testing.assert_almost_equal(rms(wave), 0.7071067811865476, decimal=5)
    np.testing.assert_almost_equal(rms(offsetWave), 2.1213203435596424, decimal=5)
    np.testing.assert_almost_equal(rms(noiseWave), 0.7128245483240299, decimal=5)


def test_ecdf():
    np.testing.assert_almost_equal(ecdf(const0), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(const1), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(constNeg), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(constF), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(lin), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(lin0), (0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5))
    np.testing.assert_almost_equal(ecdf(wave), (0.001, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.009,0.01))
    np.testing.assert_almost_equal(ecdf(offsetWave), (0.001, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.009,0.01))
    np.testing.assert_almost_equal(ecdf(noiseWave), (0.001, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.009, 0.01))

def test_ecdf_percentile():
    np.testing.assert_almost_equal(ecdf_percentile(const0), 0)
    np.testing.assert_almost_equal(ecdf_percentile(const1), 1)
    np.testing.assert_almost_equal(ecdf_percentile(constNeg), -1)
    np.testing.assert_almost_equal(ecdf_percentile(constF), 2.5)
    np.testing.assert_almost_equal(ecdf_percentile(lin), (3, 15))
    np.testing.assert_almost_equal(ecdf_percentile(lin0), (-6.8421053, 5.7894737))
    np.testing.assert_almost_equal(ecdf_percentile(wave), (-0.809017, 0.809017))
    np.testing.assert_almost_equal(ecdf_percentile(offsetWave), (1.1909830056250523, 2.809016994374947))
    np.testing.assert_almost_equal(ecdf_percentile(noiseWave), (-0.8095410722491809, 0.796916231269631))


def test_ecdf_slope():
    np.testing.assert_almost_equal(ecdf_slope(const0), np.inf)
    np.testing.assert_almost_equal(ecdf_slope(const1), np.inf)
    np.testing.assert_almost_equal(ecdf_slope(constNeg), np.inf)
    np.testing.assert_almost_equal(ecdf_slope(constF), np.inf)
    np.testing.assert_almost_equal(ecdf_slope(lin), 0.05)
    np.testing.assert_almost_equal(ecdf_slope(lin0), 0.0475)
    np.testing.assert_almost_equal(ecdf_slope(wave), 0.3535534)
    np.testing.assert_almost_equal(ecdf_slope(offsetWave), 0.3535534)
    np.testing.assert_almost_equal(ecdf_slope(noiseWave), 0.3594095)


def test_ecdf_percentile_count():
    np.testing.assert_almost_equal(ecdf_percentile_count(const0), (0, 0))
    np.testing.assert_almost_equal(ecdf_percentile_count(const1), (1, 1))
    np.testing.assert_almost_equal(ecdf_percentile_count(constNeg), (-1, -1))
    np.testing.assert_almost_equal(ecdf_percentile_count(constF), (2.5, 2.5))
    np.testing.assert_almost_equal(ecdf_percentile_count(lin), (4, 16))
    np.testing.assert_almost_equal(ecdf_percentile_count(lin0), (4, 16))
    np.testing.assert_almost_equal(ecdf_percentile_count(wave), (200, 800))
    np.testing.assert_almost_equal(ecdf_percentile_count(offsetWave), (200, 800))
    np.testing.assert_almost_equal(ecdf_percentile_count(noiseWave), (200, 800))


# ################################################ TEMPORAL FEATURES ################################################# #
def test_distance():
    np.testing.assert_almost_equal(distance(const0), 19.0)
    np.testing.assert_almost_equal(distance(const1), 19.0)
    np.testing.assert_almost_equal(distance(constNeg), 19.0)
    np.testing.assert_almost_equal(distance(constF), 19.0)
    np.testing.assert_almost_equal(distance(lin), 26.87005768508881)
    np.testing.assert_almost_equal(distance(lin0), 27.586228448267438, decimal=5)
    np.testing.assert_almost_equal(distance(wave), 999.2461809866238, decimal=5)
    np.testing.assert_almost_equal(distance(offsetWave), 999.2461809866238, decimal=5)
    np.testing.assert_almost_equal(distance(noiseWave), 1007.8711901383033, decimal=5)


def test_minpeaks():
    np.testing.assert_almost_equal(minpeaks(const0), 0.0)
    np.testing.assert_almost_equal(minpeaks(const1), 0.0)
    np.testing.assert_almost_equal(minpeaks(constNeg), 0.0)
    np.testing.assert_almost_equal(minpeaks(constF), 0.0)
    np.testing.assert_almost_equal(minpeaks(lin), 0.0)
    np.testing.assert_almost_equal(minpeaks(lin0), 0.0, decimal=5)
    np.testing.assert_almost_equal(minpeaks(wave), 5, decimal=5)
    np.testing.assert_almost_equal(minpeaks(offsetWave), 5, decimal=5)
    np.testing.assert_almost_equal(minpeaks(noiseWave), 323, decimal=5)


def test_maxpeaks():
    np.testing.assert_almost_equal(maxpeaks(const0), 0.0)
    np.testing.assert_almost_equal(maxpeaks(const1), 0.0)
    np.testing.assert_almost_equal(maxpeaks(constNeg), 0.0)
    np.testing.assert_almost_equal(maxpeaks(constF), 0.0)
    np.testing.assert_almost_equal(maxpeaks(lin), 0.0)
    np.testing.assert_almost_equal(maxpeaks(lin0), 0.0, decimal=5)
    np.testing.assert_almost_equal(maxpeaks(wave), 5, decimal=5)
    np.testing.assert_almost_equal(maxpeaks(offsetWave), 5, decimal=5)
    np.testing.assert_almost_equal(maxpeaks(noiseWave), 322, decimal=5)


def test_centroid():
    np.testing.assert_almost_equal(calc_centroid(const0, Fs), 0.0)
    np.testing.assert_almost_equal(calc_centroid(const1, Fs), 0.009499999999999998)
    np.testing.assert_almost_equal(calc_centroid(constNeg, Fs), 0.009499999999999998)
    np.testing.assert_almost_equal(calc_centroid(constF, Fs), 0.0095)
    np.testing.assert_almost_equal(calc_centroid(lin, Fs), 0.014615384615384615)
    np.testing.assert_almost_equal(calc_centroid(lin0, Fs), 0.0095, decimal=5)
    np.testing.assert_almost_equal(calc_centroid(wave, Fs), 0.5000000000000001, decimal=5)
    np.testing.assert_almost_equal(calc_centroid(offsetWave, Fs), 0.47126367059427926, decimal=5)
    np.testing.assert_almost_equal(calc_centroid(noiseWave, Fs), 0.4996034303128802, decimal=5)


def test_mean_diff():
    np.testing.assert_almost_equal(mean_diff(const0), 0.0)
    np.testing.assert_almost_equal(mean_diff(const1), 0.0)
    np.testing.assert_almost_equal(mean_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(mean_diff(constF), 0.0)
    np.testing.assert_almost_equal(mean_diff(lin), 1.0)
    np.testing.assert_almost_equal(mean_diff(lin0), 1.0526315789473684, decimal=5)
    np.testing.assert_almost_equal(mean_diff(wave), -3.1442201279407477e-05, decimal=5)
    np.testing.assert_almost_equal(mean_diff(offsetWave), -3.1442201279407036e-05, decimal=5)
    np.testing.assert_almost_equal(mean_diff(noiseWave), -0.00010042477181949707, decimal=5)


def test_median_diff():
    np.testing.assert_almost_equal(median_diff(const0), 0.0)
    np.testing.assert_almost_equal(median_diff(const1), 0.0)
    np.testing.assert_almost_equal(median_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(median_diff(constF), 0.0)
    np.testing.assert_almost_equal(median_diff(lin), 1.0)
    np.testing.assert_almost_equal(median_diff(lin0), 1.0526315789473684, decimal=5)
    np.testing.assert_almost_equal(median_diff(wave), -0.0004934396342684, decimal=5)
    np.testing.assert_almost_equal(median_diff(offsetWave), -0.0004934396342681779, decimal=5)
    np.testing.assert_almost_equal(median_diff(noiseWave), -0.004174819648320949, decimal=5)


def test_calc_mean_abs_diff():
    np.testing.assert_almost_equal(mean_abs_diff(const0), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(const1), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(constF), 0.0)
    np.testing.assert_almost_equal(mean_abs_diff(lin), 1.0)
    np.testing.assert_almost_equal(mean_abs_diff(lin0), 1.0526315789473684, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(wave), 0.019988577818740614, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(offsetWave), 0.019988577818740614, decimal=5)
    np.testing.assert_almost_equal(mean_abs_diff(noiseWave), 0.10700252903161508, decimal=5)


def test_median_abs_diff():
    np.testing.assert_almost_equal(median_abs_diff(const0), 0.0)
    np.testing.assert_almost_equal(median_abs_diff(const1), 0.0)
    np.testing.assert_almost_equal(median_abs_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(median_abs_diff(constF), 0.0)
    np.testing.assert_almost_equal(median_abs_diff(lin), 1.0)
    np.testing.assert_almost_equal(median_abs_diff(lin0), 1.0526315789473681, decimal=5)
    np.testing.assert_almost_equal(median_abs_diff(wave), 0.0218618462348652, decimal=5)
    np.testing.assert_almost_equal(median_abs_diff(offsetWave), 0.021861846234865645, decimal=5)
    np.testing.assert_almost_equal(median_abs_diff(noiseWave), 0.08958750592592835, decimal=5)


def test_sum_abs_diff():
    np.testing.assert_almost_equal(sum_abs_diff(const0), 0.0)
    np.testing.assert_almost_equal(sum_abs_diff(const1), 0.0)
    np.testing.assert_almost_equal(sum_abs_diff(constNeg), 0.0)
    np.testing.assert_almost_equal(sum_abs_diff(constF), 0.0)
    np.testing.assert_almost_equal(sum_abs_diff(lin), 19)
    np.testing.assert_almost_equal(sum_abs_diff(lin0), 20.0, decimal=5)
    np.testing.assert_almost_equal(sum_abs_diff(wave), 19.968589240921872, decimal=5)
    np.testing.assert_almost_equal(sum_abs_diff(offsetWave), 19.968589240921872, decimal=5)
    np.testing.assert_almost_equal(sum_abs_diff(noiseWave), 106.89552650258346, decimal=5)


def test_zerocross():
    np.testing.assert_almost_equal(zero_cross(const0), 0.0)
    np.testing.assert_almost_equal(zero_cross(const1), 0.0)
    np.testing.assert_almost_equal(zero_cross(constNeg), 0.0)
    np.testing.assert_almost_equal(zero_cross(constF), 0.0)
    np.testing.assert_almost_equal(zero_cross(lin), 1.0)
    np.testing.assert_almost_equal(zero_cross(lin0), 1.0, decimal=5)
    np.testing.assert_almost_equal(zero_cross(wave), 10, decimal=5)
    np.testing.assert_almost_equal(zero_cross(offsetWave), 0.0, decimal=5)
    np.testing.assert_almost_equal(zero_cross(noiseWave), 38, decimal=5)


def test_autocorr():
    np.testing.assert_almost_equal(autocorr(const0), 0.0)
    np.testing.assert_almost_equal(autocorr(const1), 20.0)
    np.testing.assert_almost_equal(autocorr(constNeg), 20.0)
    np.testing.assert_almost_equal(autocorr(constF), 125.0)
    np.testing.assert_almost_equal(autocorr(lin), 2470.0)
    np.testing.assert_almost_equal(autocorr(lin0), 736.8421052631579, decimal=0)
    np.testing.assert_almost_equal(autocorr(wave), 500.5, decimal=0)
    np.testing.assert_almost_equal(autocorr(offsetWave), 4500.0, decimal=0)
    np.testing.assert_almost_equal(autocorr(noiseWave), 508.6149018530489, decimal=0)


def test_auc():
    np.testing.assert_almost_equal(auc(const0, Fs), 0.0)
    np.testing.assert_almost_equal(auc(const1, Fs), 9.518999999999998)
    np.testing.assert_almost_equal(auc(constNeg, Fs), -9.518999999999998)
    np.testing.assert_almost_equal(auc(constF, Fs), 23.797500000000003)
    np.testing.assert_almost_equal(auc(lin, Fs), 95.171)
    np.testing.assert_almost_equal(auc(lin0, Fs), 4.989999999999997)
    np.testing.assert_almost_equal(auc(wave, Fs), 3.1410759074645966e-05)
    np.testing.assert_almost_equal(auc(offsetWave, Fs), 1000.998031410759)
    np.testing.assert_almost_equal(auc(noiseWave, Fs), -0.7958996038449087)


def test_abs_energy():
    np.testing.assert_almost_equal(abs_energy(const0), 0.0)
    np.testing.assert_almost_equal(abs_energy(const1), 20.0)
    np.testing.assert_almost_equal(abs_energy(constNeg), 20.0)
    np.testing.assert_almost_equal(abs_energy(constF), 125.0)
    np.testing.assert_almost_equal(abs_energy(lin), 2470)
    np.testing.assert_almost_equal(abs_energy(lin0), 736.8421052631579)
    np.testing.assert_almost_equal(abs_energy(wave), 500.0)
    np.testing.assert_almost_equal(abs_energy(offsetWave), 4500.0)
    np.testing.assert_almost_equal(abs_energy(noiseWave), 508.11883669335725)


def test_pk_pk_distance():
    np.testing.assert_almost_equal(pk_pk_distance(const0), 0.0)
    np.testing.assert_almost_equal(pk_pk_distance(const1), 0.0)
    np.testing.assert_almost_equal(pk_pk_distance(constNeg), 0.0)
    np.testing.assert_almost_equal(pk_pk_distance(constF), 0.0)
    np.testing.assert_almost_equal(pk_pk_distance(lin), 19)
    np.testing.assert_almost_equal(pk_pk_distance(lin0), 20.0)
    np.testing.assert_almost_equal(pk_pk_distance(wave), 2.0)
    np.testing.assert_almost_equal(pk_pk_distance(offsetWave), 2.0)
    np.testing.assert_almost_equal(pk_pk_distance(noiseWave), 2.4800109800001993)


def test_slope():
    np.testing.assert_almost_equal(slope(const0), 0.0)
    np.testing.assert_almost_equal(slope(const1), -8.935559365603017e-18)
    np.testing.assert_almost_equal(slope(constNeg), 8.935559365603017e-18)
    np.testing.assert_almost_equal(slope(constF), 1.7871118731206033e-17)
    np.testing.assert_almost_equal(slope(lin), 1.0)
    np.testing.assert_almost_equal(slope(lin0), 1.0526315789473686)
    np.testing.assert_almost_equal(slope(wave), -0.0003819408289180587)
    np.testing.assert_almost_equal(slope(offsetWave), -0.00038194082891805853)
    np.testing.assert_almost_equal(slope(noiseWave), -0.00040205425841671337)


def test_entropy():
    np.testing.assert_almost_equal(entropy(const0), 0.0)
    np.testing.assert_almost_equal(entropy(const1), 0.0)
    np.testing.assert_almost_equal(entropy(constNeg), 0.0)
    np.testing.assert_almost_equal(entropy(constF), 0.0)
    np.testing.assert_almost_equal(entropy(lin), 0.994983274605318)
    np.testing.assert_almost_equal(entropy(lin0), 0.994983274605318)
    np.testing.assert_almost_equal(entropy(wave), 0.9972021515128497)
    np.testing.assert_almost_equal(entropy(offsetWave), 0.99720215151285)
    np.testing.assert_almost_equal(entropy(noiseWave), 0.9957000733996481)


# ################################################ SPECTRAL FEATURES ################################################# #
def test_max_fre():
    np.testing.assert_almost_equal(max_frequency(const0, Fs), 0.0)
    np.testing.assert_almost_equal(max_frequency(const1, Fs), 0.0)
    np.testing.assert_almost_equal(max_frequency(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(max_frequency(constF, Fs), 0.0)
    np.testing.assert_almost_equal(max_frequency(lin, Fs), 444.44444444444446)
    np.testing.assert_almost_equal(max_frequency(lin0, Fs), 500.0, decimal=5)
    np.testing.assert_almost_equal(max_frequency(wave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(max_frequency(offsetWave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(max_frequency(noiseWave, Fs), 464.9298597194388, decimal=5)
    np.testing.assert_almost_equal(max_frequency(x, Fs), 344.689378757515, decimal=1)


def test_med_fre():
    np.testing.assert_almost_equal(median_frequency(const0, Fs), 0.0)
    np.testing.assert_almost_equal(median_frequency(const1, Fs), 0.0)
    np.testing.assert_almost_equal(median_frequency(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(median_frequency(constF, Fs), 0.0)
    np.testing.assert_almost_equal(median_frequency(lin, Fs), 55.55555555555556)
    np.testing.assert_almost_equal(median_frequency(lin0, Fs), 166.66666666666669, decimal=5)
    np.testing.assert_almost_equal(median_frequency(wave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(median_frequency(offsetWave, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(median_frequency(noiseWave, Fs), 146.29258517034066, decimal=5)
    np.testing.assert_almost_equal(median_frequency(x, Fs), 4.008016032064128, decimal=1)


def test_fund_fre():
    np.testing.assert_almost_equal(fundamental_frequency(const0, 1), 0.0)
    np.testing.assert_almost_equal(fundamental_frequency(const1, 1), 0.0)
    np.testing.assert_almost_equal(fundamental_frequency(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(fundamental_frequency(constF, Fs), 0.0)
    np.testing.assert_almost_equal(fundamental_frequency(lin, Fs), 0.0)
    np.testing.assert_almost_equal(fundamental_frequency(lin0, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(fundamental_frequency(wave, Fs), 5.0100200400801596, decimal=1)
    np.testing.assert_almost_equal(fundamental_frequency(offsetWave, Fs), 5.0100200400801596, decimal=1)
    np.testing.assert_almost_equal(fundamental_frequency(noiseWave, Fs), 5.0100200400801596, decimal=1)


def test_power_spec():
    np.testing.assert_almost_equal(max_power_spectrum(const0, Fs), 0.0)
    np.testing.assert_almost_equal(max_power_spectrum(const1, Fs), 0.0)
    np.testing.assert_almost_equal(max_power_spectrum(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(max_power_spectrum(constF, Fs), 0.0)
    np.testing.assert_almost_equal(max_power_spectrum(lin, Fs), 0.004621506382612649)
    np.testing.assert_almost_equal(max_power_spectrum(lin0, Fs), 0.0046215063826126525, decimal=5)
    np.testing.assert_almost_equal(max_power_spectrum(wave, Fs), 0.6666666666666667, decimal=5)
    np.testing.assert_almost_equal(max_power_spectrum(offsetWave, Fs), 0.6666666666666667, decimal=5)
    np.testing.assert_almost_equal(max_power_spectrum(noiseWave, Fs), 0.6570878541643916, decimal=5)


def test_total_energy():
    np.testing.assert_almost_equal(total_energy(const0, Fs), 0.0)
    np.testing.assert_almost_equal(total_energy(const1, Fs), 1052.6315789473686)
    np.testing.assert_almost_equal(total_energy(constNeg, Fs), 1052.6315789473686)
    np.testing.assert_almost_equal(total_energy(constF, Fs), 6578.9473684210525)
    np.testing.assert_almost_equal(total_energy(lin, Fs), 130000.0)
    np.testing.assert_almost_equal(total_energy(lin0, Fs), 38781.16343490305, decimal=5)
    np.testing.assert_almost_equal(total_energy(wave, Fs), 500.5005005005005, decimal=5)
    np.testing.assert_almost_equal(total_energy(offsetWave, Fs), 4504.504504504504, decimal=5)
    np.testing.assert_almost_equal(total_energy(noiseWave, Fs), 508.6274641575148, decimal=5)


def test_spectral_centroid():
    np.testing.assert_almost_equal(spectral_centroid(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_centroid(const1, Fs), 2.7476856540265033e-14)
    np.testing.assert_almost_equal(spectral_centroid(constNeg, Fs), 2.7476856540265033e-14)
    np.testing.assert_almost_equal(spectral_centroid(constF, Fs), 2.4504208511457478e-14)
    np.testing.assert_almost_equal(spectral_centroid(lin, Fs), 95.77382394996009)
    np.testing.assert_almost_equal(spectral_centroid(lin0, Fs), 189.7228259594313, decimal=5)
    np.testing.assert_almost_equal(spectral_centroid(wave, Fs), 5.010020040084022, decimal=5)
    np.testing.assert_almost_equal(spectral_centroid(offsetWave, Fs), 1.0020040080169172, decimal=5)
    np.testing.assert_almost_equal(spectral_centroid(noiseWave, Fs), 181.12036927310848, decimal=5)


def test_spectral_spread():
    np.testing.assert_almost_equal(spectral_spread(const0, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(const1, Fs), 2.811883163207112e-06, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(constNeg, Fs), 2.811883163207112e-06, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(constF, Fs), 2.657703172211011e-06, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(lin, Fs), 137.9288076645223, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(lin0, Fs), 140.93247375966078, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(wave, Fs), 3.585399057660381e-05, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(offsetWave, Fs), 2.004008016105514, decimal=5)
    np.testing.assert_almost_equal(spectral_spread(noiseWave, Fs), 165.6402040682083, decimal=5)


def test_spectral_skewness():
    np.testing.assert_almost_equal(spectral_skewness(const0, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(const1, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(constNeg, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(constF, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(lin, Fs), 1.5090650071326563, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(lin0, Fs), 0.8140329168647044, decimal=5)
    np.testing.assert_almost_equal(spectral_skewness(wave, Fs), 10643315.707158063, decimal=1)
    np.testing.assert_almost_equal(spectral_skewness(offsetWave, Fs), 1.5000000137542306, decimal=1)
    np.testing.assert_almost_equal(spectral_skewness(noiseWave, Fs), 0.4126776686583098, decimal=1)


def test_spectral_kurtosis():
    np.testing.assert_almost_equal(spectral_kurtosis(const0, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(const1, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(constNeg, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(constF, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(lin, Fs), 4.209140226148914, decimal=0)
    np.testing.assert_almost_equal(spectral_kurtosis(lin0, Fs), 2.4060168768515413, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(wave, Fs), 120959227206031.11, decimal=1)
    np.testing.assert_almost_equal(spectral_kurtosis(offsetWave, Fs), 3.2500028252333513, decimal=5)
    np.testing.assert_almost_equal(spectral_kurtosis(noiseWave, Fs), 1.7251592171239667, decimal=5)


def test_spectral_slope():
    np.testing.assert_almost_equal(spectral_slope(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_slope(const1, Fs), -0.0009818181818181818)
    np.testing.assert_almost_equal(spectral_slope(constNeg, Fs), -0.0009818181818181818)
    np.testing.assert_almost_equal(spectral_slope(constF, Fs), -0.0009818181818181816)
    np.testing.assert_almost_equal(spectral_slope(lin, Fs), -0.0006056882550328839)
    np.testing.assert_almost_equal(spectral_slope(lin0, Fs), -0.00023672490168659717, decimal=1)
    np.testing.assert_almost_equal(spectral_slope(wave, Fs), -2.3425149700598465e-05, decimal=5)
    np.testing.assert_almost_equal(spectral_slope(offsetWave, Fs), -2.380838323353288e-05, decimal=5)
    np.testing.assert_almost_equal(spectral_slope(noiseWave, Fs), -6.586047565550932e-06, decimal=5)


def test_spectral_decrease():
    np.testing.assert_almost_equal(spectral_decrease(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_decrease(const1, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_decrease(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_decrease(constF, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_decrease(lin, Fs), -2.255518236004341)
    np.testing.assert_almost_equal(spectral_decrease(lin0, Fs), 0.5195484076294969, decimal=5)
    np.testing.assert_almost_equal(spectral_decrease(wave, Fs), 0.19999999999999687, decimal=5)
    np.testing.assert_almost_equal(spectral_decrease(offsetWave, Fs), -26.963293719961584, decimal=5)
    np.testing.assert_almost_equal(spectral_decrease(noiseWave, Fs), 0.06053938231990085, decimal=5)


def test_spectral_roll_on():
    np.testing.assert_almost_equal(spectral_roll_on(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_on(const1, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_on(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_on(constF, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_on(lin, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_on(lin0, Fs), 55.55555555555556, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_on(wave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_on(offsetWave, Fs), 0.0, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_on(noiseWave, Fs), 5.0100200400801596, decimal=5)


def test_spectral_roll_off():
    np.testing.assert_almost_equal(spectral_roll_off(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_off(const1, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_off(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_off(constF, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_roll_off(lin, Fs), 444.44444444444446)
    np.testing.assert_almost_equal(spectral_roll_off(lin0, Fs), 500.0, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_off(wave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_off(offsetWave, Fs), 5.0100200400801596, decimal=5)
    np.testing.assert_almost_equal(spectral_roll_off(noiseWave, Fs), 464.9298597194388, decimal=5)


def test_spectral_distance():
    np.testing.assert_almost_equal(spectral_distance(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_distance(const1, Fs), -100)
    np.testing.assert_almost_equal(spectral_distance(constNeg, Fs), -100)
    np.testing.assert_almost_equal(spectral_distance(constF, Fs), -250)
    np.testing.assert_almost_equal(spectral_distance(lin, Fs), -1256.997293357373)
    np.testing.assert_almost_equal(spectral_distance(lin0, Fs), -323.15504563934024, decimal=5)
    np.testing.assert_almost_equal(spectral_distance(wave, Fs), -122500.00000000022, decimal=5)
    np.testing.assert_almost_equal(spectral_distance(offsetWave, Fs), -622500.0, decimal=5)
    np.testing.assert_almost_equal(spectral_distance(noiseWave, Fs), -124832.72310672606, decimal=5)


def test_spect_variation():
    np.testing.assert_almost_equal(spectral_variation(const0, Fs), 1.0)
    np.testing.assert_almost_equal(spectral_variation(const1, Fs), 1.0)
    np.testing.assert_almost_equal(spectral_variation(constNeg, Fs), 1.0)
    np.testing.assert_almost_equal(spectral_variation(constF, Fs), 1.0)
    np.testing.assert_almost_equal(spectral_variation(lin, Fs), 0.04096548417849766)
    np.testing.assert_almost_equal(spectral_variation(lin0, Fs), 0.39913530062615254, decimal=5)
    np.testing.assert_almost_equal(spectral_variation(wave, Fs), 0.9999999999999997, decimal=5)
    np.testing.assert_almost_equal(spectral_variation(offsetWave, Fs), 0.9999999999999999, decimal=5)
    np.testing.assert_almost_equal(spectral_variation(noiseWave, Fs), 0.9775968083533805, decimal=5)


def test_spectral_maxpeaks():
    np.testing.assert_almost_equal(spectral_maxpeaks(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_maxpeaks(const1, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_maxpeaks(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_maxpeaks(constF, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_maxpeaks(lin, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_maxpeaks(lin0, Fs), 1.0, decimal=5)
    np.testing.assert_almost_equal(spectral_maxpeaks(wave, Fs), 155, decimal=0)
    np.testing.assert_almost_equal(spectral_maxpeaks(offsetWave, Fs), 158, decimal=1)
    np.testing.assert_almost_equal(spectral_maxpeaks(noiseWave, Fs), 172.0, decimal=1)


def test_human_range_energy():
    np.testing.assert_almost_equal(human_range_energy(const0, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(const1, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(constF, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(lin, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(lin0, Fs), 0.0)
    np.testing.assert_almost_equal(human_range_energy(wave, Fs), 2.838300923247935e-33)
    np.testing.assert_almost_equal(human_range_energy(offsetWave, Fs), 1.6194431630448383e-33)
    np.testing.assert_almost_equal(human_range_energy(noiseWave, Fs), 4.5026865350839304e-05)


def test_mfcc():
    np.testing.assert_almost_equal(mfcc(const0, Fs), (-1e-08, -2.5654632210061364e-08, -4.099058125255727e-08,
                                                      -5.56956514302075e-08, -6.947048992011573e-08,
                                                      -8.203468073398136e-08,
                                                      -9.313245317896842e-08, -1.0253788861142992e-07,
                                                      -1.1005951948899701e-07,
                                                      -1.1554422709759472e-07, -1.1888035860690259e-07,
                                                      -1.2000000000000002e-07))
    np.testing.assert_almost_equal(mfcc(const1, Fs), (0.14096637144714785, 0.4029720554090289, 0.2377457745400458,
                                                      0.9307791929462678, -0.8138023913445843, -0.36127671623673,
                                                      0.17779314470940918, 1.5842014538963525, -5.868875380858009,
                                                      -1.3484207382203723, -1.5899059472962034, 2.9774371742123975))
    np.testing.assert_almost_equal(mfcc(constNeg, Fs), (0.14096637144714785, 0.4029720554090289, 0.2377457745400458,
                                                        0.9307791929462678, -0.8138023913445843, -0.36127671623673,
                                                        0.17779314470940918, 1.5842014538963525, -5.868875380858009,
                                                        -1.3484207382203723, -1.5899059472962034, 2.9774371742123975))
    np.testing.assert_almost_equal(mfcc(constF, Fs), (0.1409663714471363, 0.40297205540906766, 0.23774577454002216,
                                                      0.9307791929463864, -0.8138023913445535, -0.3612767162368284,
                                                      0.17779314470931407, 1.584201453896316, -5.868875380858139,
                                                      -1.3484207382203004, -1.589905947296293, 2.977437174212552))
    np.testing.assert_almost_equal(mfcc(lin, Fs), (63.41077963677539, 42.33256774689686, 22.945623346731722,
                                                   -9.267967765468333, -30.918618746635172, -69.45624761250505,
                                                   -81.74881720705784, -112.32234611356338, -127.73335353282954,
                                                   -145.3505024599537, -152.08439229251312, -170.61228411241296))
    np.testing.assert_almost_equal(mfcc(lin0, Fs), (4.472854975902669, 9.303621966161266, 12.815317252229947,
                                                    12.65260020301481, 9.763110307405048, 3.627814979708572,
                                                    1.0051648150842092, -8.07514557618858, -24.79987026383853,
                                                    -36.55749714126207, -49.060094200797785, -61.45654150658956))
    np.testing.assert_almost_equal(mfcc(wave, Fs), (115.31298449242963, -23.978080415791883, 64.49711308839377,
                                                    -70.83883973188331, -17.4881594184545, -122.5191336465161,
                                                    -89.73379214517978, -164.5583844690884, -153.29482394321641,
                                                    -204.0607944643521, -189.9059214788022, -219.38937674972897))
    np.testing.assert_almost_equal(mfcc(offsetWave, Fs), (0.02803261518615674, 0.21714705316418328, 0.4010268706527706,
                                                          1.0741653432632032, -0.26756380975236493,
                                                          -0.06446520044381611, 1.2229170142535633, 2.2173729990650166,
                                                          -5.161787305125577, -1.777027230578585, -2.2267834681371506,
                                                          1.266610194040295))
    np.testing.assert_almost_equal(mfcc(noiseWave, Fs), (-59.93874366630627, -20.646010360067542, -5.9381521505819,
                                                         13.868391975194648, 65.73380784148053, 67.65563377433688,
                                                         35.223042940942214, 73.01746718829553, 137.50395589362876,
                                                         111.61718917042731, 82.69709467796633, 110.67135918512074))


def test_power_bandwidth():
    np.testing.assert_almost_equal(power_bandwidth(const0, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(const1, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(constF, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(lin, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(lin0, Fs), 0.0)
    np.testing.assert_almost_equal(power_bandwidth(wave, Fs), 2.0)
    np.testing.assert_almost_equal(power_bandwidth(offsetWave, Fs), 2.0)
    np.testing.assert_almost_equal(power_bandwidth(noiseWave, Fs), 2.0)


def test_fft_mean_coeff():
    np.testing.assert_almost_equal(fft_mean_coeff(const0, Fs, nfreq=10),
                                   (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(fft_mean_coeff(const1, Fs, nfreq=10),
                                   (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(fft_mean_coeff(constNeg, Fs, nfreq=10),
                                   (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(fft_mean_coeff(constF, Fs, nfreq=10),
                                   (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(fft_mean_coeff(lin, Fs, nfreq=10), (0.00408221375370652, 0.29732082717207287,
                                                                       0.04400486791011177, 0.006686945426272411,
                                                                       0.00027732608206304087, 0.0003337183893114616,
                                                                       0.0008722727267959805, 0.0007221373313148659,
                                                                       0.00024061479410220662, 2.1097101108186473e-07))
    np.testing.assert_almost_equal(fft_mean_coeff(lin0, Fs, nfreq=10), (0.004523228535962903, 0.3294413597474491,
                                                                        0.04875885641009613, 0.007409357813044217,
                                                                        0.00030728651752137475, 0.0003697710684891545,
                                                                        0.0009665071765052403, 0.0008001521676618994,
                                                                        0.00026660919014094884, 2.337628931654879e-07))
    np.testing.assert_almost_equal(fft_mean_coeff(wave, Fs, nfreq=10), (2.0234880089914443e-06, 0.0001448004568848076,
                                                                        2.1047578415647817e-05, 3.2022732210152474e-06,
                                                                        1.52158292419209e-07, 1.7741879185514087e-07,
                                                                        4.2795757073284126e-07, 3.5003942541628605e-07,
                                                                        1.1626895252132188e-07, 1.6727906953620535e-10))
    np.testing.assert_almost_equal(fft_mean_coeff(offsetWave, Fs, nfreq=10), (2.0234880089914642e-06,
                                                                              0.00014480045688480763,
                                                                              2.104757841564781e-05,
                                                                              3.2022732210152483e-06,
                                                                              1.5215829241920897e-07,
                                                                              1.7741879185514156e-07,
                                                                              4.27957570732841e-07,
                                                                              3.500394254162859e-07,
                                                                              1.1626895252132173e-07,
                                                                              1.6727906953620255e-10))
    np.testing.assert_almost_equal(fft_mean_coeff(noiseWave, Fs, nfreq=10), (3.2947755935395495e-06,
                                                                             0.00014466702099241778,
                                                                             3.838265852158549e-05,
                                                                             1.6729032217627548e-05,
                                                                             1.6879950037320804e-05,
                                                                             1.571169205601392e-05,
                                                                             1.679718723715948e-05,
                                                                             1.810371503556574e-05,
                                                                             2.0106126483830693e-05,
                                                                             8.91285109135437e-06))


def test_lpcc():
    np.testing.assert_almost_equal(lpcc(const0), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    np.testing.assert_almost_equal(lpcc(const1), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    np.testing.assert_almost_equal(lpcc(constNeg), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    np.testing.assert_almost_equal(lpcc(constF), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    np.testing.assert_almost_equal(lpcc(lin), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    np.testing.assert_almost_equal(lpcc(lin0), (0.017793342850434657, 0.12419699587050197, 0.17985773867565555,
                                                0.13749027713829948, 0.14521059821841656, 0.14362411136332903,
                                                0.14403924127165643, 0.14362411136332903, 0.14521059821841656,
                                                0.13749027713829948, 0.17985773867565555, 0.12419699587050197))
    np.testing.assert_almost_equal(lpcc(wave), (8.08705689884851e-07, 0.10193422882411193, 0.0051922525746904875,
                                                0.0003496693593067946, 2.355214618130234e-05, 1.2419899263690914e-06,
                                                3.091008802744081e-06, 1.2419899263690914e-06, 2.355214618130234e-05,
                                                0.0003496693593067946, 0.0051922525746904875, 0.10193422882411193))
    np.testing.assert_almost_equal(lpcc(offsetWave), (8.087054868870942e-07, 0.10193422882503231, 0.005192252575236202,
                                                      0.0003496693583308415, 2.3552147454092374e-05,
                                                      1.241991615337501e-06, 3.0910069449505212e-06,
                                                      1.241991615337501e-06, 2.3552147454092374e-05,
                                                      0.0003496693583308415, 0.005192252575236202, 0.10193422882503231))
    np.testing.assert_almost_equal(lpcc(noiseWave), (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))


def test_spectral_entropy():
    np.testing.assert_almost_equal(spectral_entropy(const0, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_entropy(const1, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_entropy(constNeg, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_entropy(constF, Fs), 0.0)
    np.testing.assert_almost_equal(spectral_entropy(lin, Fs), 0.6006757398806453)
    np.testing.assert_almost_equal(spectral_entropy(lin0, Fs), 0.57319032538303)
    np.testing.assert_almost_equal(spectral_entropy(wave, Fs), 1.5228376718814352e-29)
    np.testing.assert_almost_equal(spectral_entropy(offsetWave, Fs), 1.783049297437309e-29)
    np.testing.assert_almost_equal(spectral_entropy(noiseWave, Fs), 0.030107186831275425)


def test_wavelet_entropy():
    np.testing.assert_almost_equal(wavelet_entropy(const0), 0.0)
    np.testing.assert_almost_equal(wavelet_entropy(const1), 1.9188378548746368)
    np.testing.assert_almost_equal(wavelet_entropy(constNeg), 1.9188378548746368)
    np.testing.assert_almost_equal(wavelet_entropy(constF), 1.9188378548746368)
    np.testing.assert_almost_equal(wavelet_entropy(lin), 1.9648440772467513)
    np.testing.assert_almost_equal(wavelet_entropy(lin0), 2.0713919678725117)
    np.testing.assert_almost_equal(wavelet_entropy(wave), 1.7277528462213683)
    np.testing.assert_almost_equal(wavelet_entropy(offsetWave), 1.7965939302139549)
    np.testing.assert_almost_equal(wavelet_entropy(noiseWave), 2.0467527462416153)


def test_wavelet_abs_mean():
    np.testing.assert_almost_equal(wavelet_abs_mean(const0), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(wavelet_abs_mean(const1),
                                   (0.081894185676901, 0.24260084511769256, 0.4653470776794248,
                                    0.8500400580778283, 1.3602249381214044, 1.8378460432593602,
                                    2.2080039502231164, 2.4676456085810874, 2.638131856418627))
    np.testing.assert_almost_equal(wavelet_abs_mean(constNeg),
                                   (0.081894185676901, 0.24260084511769256, 0.4653470776794248,
                                    0.8500400580778283, 1.3602249381214044, 1.8378460432593602,
                                    2.2080039502231164, 2.4676456085810874, 2.638131856418627))
    np.testing.assert_almost_equal(wavelet_abs_mean(constF),
                                   (0.20473546419225214, 0.6065021127942314, 1.1633676941985622,
                                    2.1251001451945712, 3.4005623453035114, 4.5946151081484015,
                                    5.5200098755577915, 6.169114021452717, 6.5953296410465665))
    np.testing.assert_almost_equal(wavelet_abs_mean(lin), (0.7370509925842613, 2.183416725919023, 4.1974435700809565,
                                                           7.744819422931153, 12.504051331233388, 16.982183932901865,
                                                           20.46332353598833, 22.91143100556329, 24.52363151471446))
    np.testing.assert_almost_equal(wavelet_abs_mean(lin0),
                                   (0.0430987066803135, 0.12767505547269026, 0.23510912407745171,
                                    0.3479590829560181, 0.4400900851788993, 0.5024773453284851,
                                    0.5396989380329178, 0.5591602904810937, 0.5669696013289379))
    np.testing.assert_almost_equal(wavelet_abs_mean(wave), (5.138703105035948e-05, 0.00015178141653400073,
                                                            0.00027925117450851024, 0.0004278724786267016,
                                                            0.0005932191214607947, 0.0007717034331954587,
                                                            0.0009601854175466062, 0.0011557903088208192,
                                                            0.0013558175034366186))
    np.testing.assert_almost_equal(wavelet_abs_mean(offsetWave), (0.0032504208945027323, 0.009623752088931016,
                                                                  0.017761411181034453, 0.027372614777691914,
                                                                  0.03826512918833778, 0.050306487368868114,
                                                                  0.06339897203822373, 0.07746693331944604,
                                                                  0.09244971907566273))
    np.testing.assert_almost_equal(wavelet_abs_mean(noiseWave), (4.631139377647647e-05, 7.893225282164063e-05,
                                                                 0.00033257747958655794, 0.0005792253883615155,
                                                                 0.0007699898255271558, 0.0009106252575513913,
                                                                 0.0010387197644970154, 0.0011789334866018457,
                                                                 0.0013341945911985783))


def test_wavelet_std():
    np.testing.assert_almost_equal(wavelet_std(const0), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(wavelet_std(const1), (0.1767186264889806, 0.28069306259219023, 0.3235061868750311,
                                                         0.3115893726751135, 0.31446140614407014, 0.3582016825631658,
                                                         0.4133090941627322, 0.4598585090675407, 0.4935514064162697))
    np.testing.assert_almost_equal(wavelet_std(constNeg), (0.1767186264889806, 0.28069306259219023, 0.3235061868750311,
                                                           0.3115893726751135, 0.31446140614407014, 0.3582016825631658,
                                                           0.4133090941627322, 0.4598585090675407, 0.4935514064162697))
    np.testing.assert_almost_equal(wavelet_std(constF), (0.44179656622245145, 0.7017326564804757, 0.8087654671875778,
                                                         0.7789734316877838, 0.7861535153601755, 0.8955042064079146,
                                                         1.0332727354068305, 1.1496462726688517, 1.2338785160406742))
    np.testing.assert_almost_equal(wavelet_std(lin), (2.721791561180164, 5.325234998810811, 8.137581399111415,
                                                      10.529795250703716, 11.836525442245224, 12.296195571788726,
                                                      12.315744378517108, 12.135259348389042, 11.869294506387352))
    np.testing.assert_almost_equal(wavelet_std(lin0), (2.239406940011677, 4.7878443746478245, 7.797954379287043,
                                                       10.418506686200207, 11.746946049852674, 12.045972295386465,
                                                       11.828477896749822, 11.408150997410496, 10.932763618021895))
    np.testing.assert_almost_equal(wavelet_std(wave), (0.001939366875349316, 0.009733675496927717, 0.025635801097107388,
                                                       0.05125305898778544, 0.08783649118731567, 0.13636963970273208,
                                                       0.197613166916789, 0.2721306670702481, 0.360305525758368))
    np.testing.assert_almost_equal(wavelet_std(offsetWave),
                                   (0.05459142980660159, 0.10410347082332229, 0.155831467554863,
                                    0.2101395066938644, 0.268489203478025, 0.33264452641566,
                                    0.4044076212671741, 0.4854392072251105, 0.5771385517659353))
    np.testing.assert_almost_equal(wavelet_std(noiseWave),
                                   (0.08974931069698587, 0.09625674025798765, 0.10445386849293256,
                                    0.11395751571203461, 0.13232763520967267, 0.16659967802754122,
                                    0.2187573594673847, 0.2877270278501564, 0.3722670641715661))


def test_wavelet_var():
    np.testing.assert_almost_equal(wavelet_var(const0), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(wavelet_var(const1), (0.031229472948151833, 0.07878859538738324, 0.10465625294642253,
                                                         0.09708793716407076, 0.09888597595410582, 0.128308445391083,
                                                         0.17082440731761822, 0.21146984836182142, 0.24359299077547786))
    np.testing.assert_almost_equal(wavelet_var(constNeg),
                                   (0.031229472948151833, 0.07878859538738324, 0.10465625294642253,
                                    0.09708793716407076, 0.09888597595410582, 0.128308445391083,
                                    0.17082440731761822, 0.21146984836182142, 0.24359299077547786))
    np.testing.assert_almost_equal(wavelet_var(constF), (0.19518420592594893, 0.49242872117114533, 0.654101580915141,
                                                         0.6067996072754422, 0.6180373497131617, 0.8019277836942689,
                                                         1.0676525457351138, 1.3216865522613839, 1.5224561923467361))
    np.testing.assert_almost_equal(wavelet_var(lin), (7.408149302511555, 28.35812779255958, 66.22023102716409,
                                                      110.87658802174253, 140.10333454491848, 151.19642553967668,
                                                      151.67755959697575, 147.26451945266362, 140.88015207935698))
    np.testing.assert_almost_equal(wavelet_var(lin0), (5.014943442972464, 22.923453755846815, 60.808092501441976,
                                                       108.5452815703984, 137.99074149814933, 145.10544854121827,
                                                       139.91288935389912, 130.1459091797181, 119.5253203275432))
    np.testing.assert_almost_equal(wavelet_var(wave),
                                   (3.761143877202169e-06, 9.474443867949103e-05, 0.0006571942978904524,
                                    0.0026268760556054137, 0.007715249184099382, 0.018596678632652963,
                                    0.03905096373888271, 0.07405509996009821, 0.12982007189201397))
    np.testing.assert_almost_equal(wavelet_var(offsetWave),
                                   (0.0029802242083291084, 0.010837532637462314, 0.02428344628030232,
                                    0.044158612273540676, 0.07208645238426431, 0.11065238095429873,
                                    0.16354552413897414, 0.2356512239113438, 0.33308890793448115))
    np.testing.assert_almost_equal(wavelet_var(noiseWave),
                                   (0.008054938770584103, 0.0092653600450937, 0.01091061064313885,
                                    0.012986315387258616, 0.017510603040184203, 0.027755452718880403,
                                    0.04785478232114257, 0.08278684255548469, 0.1385827670669169))


def test_wavelet_entropy():
    np.testing.assert_almost_equal(wavelet_energy(const0), (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    np.testing.assert_almost_equal(wavelet_energy(const1), (0.19477199643643478, 0.3710037269882903, 0.56674875884399,
                                                            0.9053485723747671, 1.3961009484422982, 1.8724279756816202,
                                                            2.2463539016634275, 2.510128422593423, 2.683902509896041))
    np.testing.assert_almost_equal(wavelet_energy(constNeg), (0.19477199643643478, 0.3710037269882903, 0.56674875884399,
                                                              0.9053485723747671, 1.3961009484422982,
                                                              1.8724279756816202,
                                                              2.2463539016634275, 2.510128422593423, 2.683902509896041))
    np.testing.assert_almost_equal(wavelet_energy(constF), (0.48692999109108687, 0.9275093174707258, 1.4168718971099752,
                                                            2.263371430936918, 3.4902523711057456, 4.6810699392040505,
                                                            5.615884754158569, 6.275321056483556, 6.709756274740101))
    np.testing.assert_almost_equal(wavelet_energy(lin), (2.819821531264169, 5.7554701277638936, 9.156350995411767,
                                                         13.071297407509103, 17.21785800380053, 20.966425462405052,
                                                         23.883575313078858, 25.926785187819767, 27.244974853151422))
    np.testing.assert_almost_equal(wavelet_energy(lin0), (2.2398216316238173, 4.789546395603321, 7.8014978562880115,
                                                          10.424315665491429, 11.75518697346929, 12.056447736534448,
                                                          11.84078393931808, 11.421846147193937, 10.947455177180416))
    np.testing.assert_almost_equal(wavelet_energy(wave),
                                   (0.0019400475520363772, 0.00973485882167256, 0.025637321995655413,
                                    0.051254844946242696, 0.08783849436907175, 0.13637182318514984,
                                    0.19761549963228792, 0.2721331214889804, 0.3603080766970352))
    np.testing.assert_almost_equal(wavelet_energy(offsetWave),
                                   (0.054688110630378595, 0.10454735406375197, 0.15684040935755078,
                                    0.21191477606176637, 0.27120227229148447, 0.3364269959823273,
                                    0.4093469845918956, 0.49158147815928066, 0.584496243351187))
    np.testing.assert_almost_equal(wavelet_energy(noiseWave),
                                   (0.08974932264551803, 0.09625677262091348, 0.10445439794914707,
                                    0.11395898775133596, 0.13232987540429264, 0.16660216672432593,
                                    0.2187598255162308, 0.2877294431226156, 0.37226945502166053))


run_module_suite()



import importlib
import inspect
import json
import os
import sys
import warnings
from inspect import getmembers, isfunction

from tsfel.feature_extraction.features_settings import load_json
from tsfel.utils.calculate_complexity import compute_complexity


def add_feature_json(features_path, json_path):
    """Adds new feature to features.json.

    Parameters
    ----------
    features_path: string
        Personal Python module directory containing new features implementation.

    json_path: string
        Personal .json file directory containing existing features from TSFEL.
        New customised features will be added to file in this directory.

    """

    sys.path.append(features_path[:-len(features_path.split(os.sep)[-1]) - 1])
    exec("import " + features_path.split(os.sep)[-1][:-3])

    # Reload module containing the new features
    importlib.reload(sys.modules[features_path.split(os.sep)[-1][:-3]])
    exec("import " + features_path.split(os.sep)[-1][:-3] + " as pymodule")

    # Functions from module containing the new features
    functions_list = [o for o in getmembers(locals()['pymodule']) if isfunction(o[1])]
    function_names = [fname[0] for fname in functions_list]

    # Check if @set_domain was declared on features module
    vset_domain = False

    for fname, f in locals()['pymodule'].__dict__.items():

        if getattr(f, "domain", None) is not None:

            vset_domain = True

            # Access to personal features.json
            feat_json = load_json(json_path)

            # Assign domain
            domain = getattr(f, "domain", None)

            # Feature specifications
            # Description
            if f.__doc__ is not None:
                descrip = f.__doc__.split("\n")[0]
            else:
                descrip = ""
            # Feature usage
            use = "yes"
            # Feature function arguments
            args_name = inspect.getfullargspec(f)[0]

            # Access feature parameters
            if args_name != "":
                # Retrieve default values of arguments
                spec = inspect.getfullargspec(f)
                defaults = dict(zip(spec.args[::-1], (spec.defaults or ())[::-1]))
                defaults.update(spec.kwonlydefaults or {})

                for p in args_name[1:]:
                    if p not in list(defaults.keys()):
                        if p is 'fs':
                            # Assigning a default value for fs if not given
                            defaults[p] = 100
                        else:
                            defaults[p] = None
                if len(defaults) == 0:
                    defaults = ""
            else:
                defaults = ""

            # Settings of new feature
            new_feature = {"description": descrip,
                           "parameters": defaults,
                           "function": fname,
                           "use": use
                           }

            # Check if domain exists
            try:
                feat_json[domain][fname] = new_feature
            except KeyError:
                feat_json[domain] = {fname: new_feature}

            # Write new feature on json file
            with open(json_path, "w") as fout:
                json.dump(feat_json, fout, indent=" ")

            # Calculate feature complexity
            compute_complexity(fname, domain, json_path, features_path=features_path)
            print('Feature '+str(fname)+' was added.')

    if vset_domain is False:
        warnings.warn('No features were added. Please declare @set_domain.', stacklevel=2)





import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    install_reqs = [
        s for s in [
            line.strip(' \n') for line in f
        ] if not s.startswith('#') and s != ''
    ]

setuptools.setup(
    name="tsfel",
    version="0.1.2",
    author="Fraunhofer Portugal",
    description="Library for time series feature extraction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fraunhoferportugal/tsfel/",
    package_data={'tsfel': ['feature_extraction/features.json', 'utils/client_secret.json']},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_reqs
)



import tsfel


FEATURES_JSON = tsfel.__path__[0] + '/feature_extraction/features.json'

settings0 = tsfel.load_json(FEATURES_JSON)

settings1 = tsfel.get_features_by_domain('statistical')

settings2 = tsfel.get_features_by_domain('temporal')

settings3 = tsfel.get_features_by_domain('spectral')

settings4 = tsfel.get_features_by_domain(None)

settings5 = tsfel.extract_sheet('Features')